from pydantic import ValidationError
from typing import List
from pathlib import Path
import yaml
import re

from visivo.parsers.yaml_ordered_dict import YamlOrderedDict


def find_line_string_start(file_path, input_dict):
    with open(file_path, "r") as file:
        file_content = file.read()

    file_lines = re.sub(r"^[ \t-]+", "", file_content, flags=re.MULTILINE).split("\n")
    input_line = re.sub(
        r"^[\s-]+", "", yaml.dump(input_dict), flags=re.MULTILINE
    ).split("\n")[0]
    for line_number, file_line in enumerate(file_lines):
        if input_line in file_line:
            return line_number + 1
    return -1


class LineValidationError(Exception):
    def __init__(self, validation_error: ValidationError, files: List[Path]):
        self.validation_error = validation_error
        self.files = files

    def get_line_message(self, error):
        if "input" in error:
            if isinstance(error["input"], YamlOrderedDict) and len(error["input"]) > 0:
                return f"  Location: {list(error['input']._key_locs.values())[0]}\n"
            if isinstance(error["input"], int) or isinstance(error["input"], str):
                input_dict = {}
                input_dict[error["loc"][-1]] = error["input"]

                for file in self.files:
                    line = find_line_string_start(file, input_dict)
                    if line >= 0:
                        return f"     Location: {file}:{line}\n\n"
        return None

    def __str__(self):
        message = f"{self.validation_error.error_count()} validation errors in {self.validation_error.title}\n\n"
        file_found = False
        for i, error in enumerate(self.validation_error.errors(), start=1):
            message = (
                message
                + f"  {i}. Object: {'.'.join(map(lambda l: str(l), error['loc']))}\n"
            )
            message = message + f"     Message: {error['msg']}\n"
            line_message = self.get_line_message(error)
            message = message + f"     The input used: ({error['input']})\n"
            if line_message:
                file_found = True
                message = message + line_message

        if file_found:
            return message
        else:
            return str(self.validation_error)



from visivo.models.base.named_model import NamedModel
from visivo.models.selector import Selector
from visivo.models.targets.target import Target
from ..models.project import Project
from ..models.base.base_model import BaseModel
from ..models.base.parent_model import ParentModel
from visivo.models.chart import Chart
from visivo.models.table import Table
from visivo.models.trace import Trace
from visivo.models.models.model import Model
from importlib.metadata import version


class Serializer:
    def __init__(self, project: Project):
        self.project = project

    def dereference(self) -> Project:
        project = self.project.model_copy(deep=True)
        project.cli_version = version("visivo")
        dag = project.dag()
        for dashboard in project.dashboards:

            def replace_item_ref(item):
                if item.chart or item.table:
                    if item.chart:
                        item.chart = ParentModel.all_descendants_of_type(
                            type=Chart, dag=dag, from_node=item
                        )[0]
                        component = item.chart
                    else:
                        item.table = ParentModel.all_descendants_of_type(
                            type=Table, dag=dag, from_node=item
                        )[0]
                        component = item.table

                    component.traces = ParentModel.all_descendants_of_type(
                        type=Trace, dag=dag, from_node=component
                    )
                    component.selector = ParentModel.all_descendants_of_type(
                        type=Selector, dag=dag, from_node=component
                    )[0]
                    for trace in component.traces:
                        trace.model = ParentModel.all_descendants_of_type(
                            type=Model, dag=dag, from_node=trace
                        )[0]
                        if hasattr(trace.model, "target"):
                            trace.model.target = ParentModel.all_descendants_of_type(
                                type=Target, dag=dag, from_node=trace.model
                            )[0]
                        if hasattr(trace.model, "models"):
                            trace.model.models = ParentModel.all_descendants_of_type(
                                type=Model, dag=dag, from_node=trace.model
                            )

            dashboard.for_each_item(replace_item_ref)

        project.charts = []
        project.traces = []
        project.tables = []
        project.models = []
        project.targets = []
        return project



from visivo.models.trace_props.bar import Bar
from visivo.models.trace_props.barpolar import Barpolar
from visivo.models.trace_props.box import Box
from visivo.models.trace_props.candlestick import Candlestick
from visivo.models.trace_props.carpet import Carpet
from visivo.models.trace_props.choropleth import Choropleth
from visivo.models.trace_props.choroplethmapbox import Choroplethmapbox
from visivo.models.trace_props.cone import Cone
from visivo.models.trace_props.contour import Contour
from visivo.models.trace_props.contourcarpet import Contourcarpet
from visivo.models.trace_props.densitymapbox import Densitymapbox
from visivo.models.trace_props.funnel import Funnel
from visivo.models.trace_props.funnelarea import Funnelarea
from visivo.models.trace_props.heatmap import Heatmap
from visivo.models.trace_props.heatmapgl import Heatmapgl
from visivo.models.trace_props.histogram import Histogram
from visivo.models.trace_props.histogram2d import Histogram2d
from visivo.models.trace_props.histogram2dcontour import Histogram2dcontour
from visivo.models.trace_props.icicle import Icicle
from visivo.models.trace_props.image import Image
from visivo.models.trace_props.indicator import Indicator
from visivo.models.trace_props.isosurface import Isosurface
from visivo.models.trace_props.mesh3d import Mesh3d
from visivo.models.trace_props.ohlc import Ohlc
from visivo.models.trace_props.parcats import Parcats
from visivo.models.trace_props.parcoords import Parcoords
from visivo.models.trace_props.pie import Pie
from visivo.models.trace_props.sankey import Sankey
from visivo.models.trace_props.scatter import Scatter
from visivo.models.trace_props.scatter3d import Scatter3d
from visivo.models.trace_props.scattercarpet import Scattercarpet
from visivo.models.trace_props.scattergeo import Scattergeo
from visivo.models.trace_props.scattergl import Scattergl
from visivo.models.trace_props.scattermapbox import Scattermapbox
from visivo.models.trace_props.scatterpolar import Scatterpolar
from visivo.models.trace_props.scatterpolargl import Scatterpolargl
from visivo.models.trace_props.scattersmith import Scattersmith
from visivo.models.trace_props.scatterternary import Scatterternary
from visivo.models.trace_props.splom import Splom
from visivo.models.trace_props.streamtube import Streamtube
from visivo.models.trace_props.sunburst import Sunburst
from visivo.models.trace_props.surface import Surface
from visivo.models.trace_props.treemap import Treemap
from visivo.models.trace_props.violin import Violin
from visivo.models.trace_props.volume import Volume
from visivo.models.trace_props.waterfall import Waterfall
from typing import Union, Annotated
from pydantic import Field

TracePropsFieldUnion = Union[
    Bar,
    Barpolar,
    Box,
    Candlestick,
    Carpet,
    Choropleth,
    Choroplethmapbox,
    Cone,
    Contour,
    Contourcarpet,
    Densitymapbox,
    Funnel,
    Funnelarea,
    Heatmap,
    Heatmapgl,
    Histogram,
    Histogram2d,
    Histogram2dcontour,
    Icicle,
    Image,
    Indicator,
    Isosurface,
    Mesh3d,
    Ohlc,
    Parcats,
    Parcoords,
    Pie,
    Sankey,
    Scatter,
    Scatter3d,
    Scattercarpet,
    Scattergeo,
    Scattergl,
    Scattermapbox,
    Scatterpolar,
    Scatterpolargl,
    Scattersmith,
    Scatterternary,
    Splom,
    Streamtube,
    Sunburst,
    Surface,
    Treemap,
    Violin,
    Volume,
    Waterfall,
]
TracePropsField = Annotated[TracePropsFieldUnion, Field(discriminator="type")]


def validate_trace_props(props):
    if "type" not in props:
        raise ValueError(f"trace_props type is required.")
    match props["type"]:
        case "bar":
            return Bar(**props)
        case "barpolar":
            return Barpolar(**props)
        case "box":
            return Box(**props)
        case "candlestick":
            return Candlestick(**props)
        case "carpet":
            return Carpet(**props)
        case "choropleth":
            return Choropleth(**props)
        case "choroplethmapbox":
            return Choroplethmapbox(**props)
        case "cone":
            return Cone(**props)
        case "contour":
            return Contour(**props)
        case "contourcarpet":
            return Contourcarpet(**props)
        case "densitymapbox":
            return Densitymapbox(**props)
        case "funnel":
            return Funnel(**props)
        case "funnelarea":
            return Funnelarea(**props)
        case "heatmap":
            return Heatmap(**props)
        case "heatmapgl":
            return Heatmapgl(**props)
        case "histogram":
            return Histogram(**props)
        case "histogram2d":
            return Histogram2d(**props)
        case "histogram2dcontour":
            return Histogram2dcontour(**props)
        case "icicle":
            return Icicle(**props)
        case "image":
            return Image(**props)
        case "indicator":
            return Indicator(**props)
        case "isosurface":
            return Isosurface(**props)
        case "mesh3d":
            return Mesh3d(**props)
        case "ohlc":
            return Ohlc(**props)
        case "parcats":
            return Parcats(**props)
        case "parcoords":
            return Parcoords(**props)
        case "pie":
            return Pie(**props)
        case "sankey":
            return Sankey(**props)
        case "scatter":
            return Scatter(**props)
        case "scatter3d":
            return Scatter3d(**props)
        case "scattercarpet":
            return Scattercarpet(**props)
        case "scattergeo":
            return Scattergeo(**props)
        case "scattergl":
            return Scattergl(**props)
        case "scattermapbox":
            return Scattermapbox(**props)
        case "scatterpolar":
            return Scatterpolar(**props)
        case "scatterpolargl":
            return Scatterpolargl(**props)
        case "scattersmith":
            return Scattersmith(**props)
        case "scatterternary":
            return Scatterternary(**props)
        case "splom":
            return Splom(**props)
        case "streamtube":
            return Streamtube(**props)
        case "sunburst":
            return Sunburst(**props)
        case "surface":
            return Surface(**props)
        case "treemap":
            return Treemap(**props)
        case "violin":
            return Violin(**props)
        case "volume":
            return Volume(**props)
        case "waterfall":
            return Waterfall(**props)
        case _:
            raise ValueError(f"{props['type']} is not a valid trace_props type.")



def run_phase(
    default_target: str,
    output_dir: str,
    working_dir: str,
    name_filter: str = None,
    run_only_changed: bool = False,
    threads: int = 8,
    soft_failure=False,
):
    from visivo.logging.logger import Logger
    from visivo.query.runner import Runner
    from visivo.commands.compile_phase import compile_phase

    project = compile_phase(
        default_target=default_target,
        working_dir=working_dir,
        output_dir=output_dir,
        name_filter=name_filter,
    )

    target_details = (
        "\n" if default_target == None else f"and default target {default_target}\n"
    )
    Logger.instance().info(f"Running project across {threads} threads" + target_details)

    runner = Runner(
        project=project,
        output_dir=output_dir,
        threads=threads,
        soft_failure=soft_failure,
        run_only_changed=run_only_changed,
        name_filter=name_filter,
    )
    runner.run()
    return runner



import click
import os
import re

from visivo.models.base.named_model import NAME_REGEX


def working_dir(function):
    click.option(
        "-w",
        "--working-dir",
        help="Directory to run the command",
        default=os.getcwd(),
    )(function)
    return function


def output_dir(function):
    click.option(
        "-o",
        "--output-dir",
        help="Directory to output results",
        default=f"{os.getcwd()}/target",
    )(function)
    return function


def dist_dir(function):
    click.option(
        "-d",
        "--dist-dir",
        help="Directory to output the distribution files",
        default=f"{os.getcwd()}/dist",
    )(function)
    return function


def name_filter(function):
    click.option(
        "-nf",
        "--name-filter",
        help="Run the command to only include the dag that includes the node with the given name",
        default=None,
    )(function)
    return function


def target(function):
    click.option(
        "-t",
        "--target",
        help="Name of the default target connection to use. This overrides the default target in the project.",
    )(function)
    return function


def alert(function):
    click.option("-a", "--alert", help="Name of the alert to use", multiple=True)(
        function
    )
    return function


def user_dir(function):
    click.option(
        "-u",
        "--user-dir",
        help="Directory containing profile",
        default=os.path.expanduser("~"),
    )(function)
    return function


def validate_stage(ctx, param, value):
    if value.strip() == "":
        raise click.BadParameter("Only whitespace is not permitted for stage name.")

    if not re.search(NAME_REGEX, value):
        raise click.BadParameter(
            "Only alphanumeric, whitespace, and '\"-_ characters permitted for stage name."
        )

    return value


def stage(function):
    click.option(
        "-s",
        "--stage",
        help="The stage of the project to deploy i.e. staging",
        callback=validate_stage,
        required=True,
    )(function)
    return function


def host(function):
    click.option(
        "-h",
        "--host",
        help="Host to deploy to",
        default=f"https://app.visivo.io",
    )(function)
    return function


def port(function):
    click.option(
        "-p",
        "--port",
        help="What port to serve on",
        default=8000,
    )(function)
    return function


def threads(function):
    click.option(
        "-th",
        "--threads",
        help="The max number of threads to use when running trace queries",
        default=8,
    )(function)
    return function



from typing import List
import warnings
from visivo.models.base.parent_model import ParentModel

from visivo.models.project import Project
from visivo.logging.logger import Logger
from time import time
from concurrent.futures import Future, ThreadPoolExecutor
import queue
from visivo.query.jobs.job import CachedFuture, Job, JobResult

from visivo.query.jobs.run_csv_script_job import jobs as csv_script_jobs
from visivo.query.jobs.run_trace_job import jobs as run_trace_jobs
from visivo.query.jobs.run_local_merge_job import jobs as run_local_merge_jobs
from visivo.query.target_job_tracker import TargetJobTracker

warnings.filterwarnings("ignore")


class Runner:
    def __init__(
        self,
        project: Project,
        output_dir: str,
        threads: int = 8,
        soft_failure=False,
        run_only_changed=False,
        name_filter: str = None,
    ):
        self.project = project
        self.output_dir = output_dir
        self.run_only_changed = run_only_changed
        self.threads = threads
        self.soft_failure = soft_failure
        self.name_filter = name_filter
        self.dag = project.dag()
        self.errors = []
        self.jobs: List[Job] = []

    def run(self):
        complete = False
        target_job_tracker = TargetJobTracker()
        start_time = time()
        self.jobs = self._all_jobs()
        with ThreadPoolExecutor(max_workers=self.threads) as executor:
            while True:
                complete = self.update_job_queue(target_job_tracker)
                if complete:
                    break

                try:
                    job = target_job_tracker.get_next_job()
                except queue.Empty:
                    continue

                if job.done():
                    pass
                elif target_job_tracker.is_accepting_job(job):
                    Logger.instance().info(job.start_message())
                    job.set_future(executor.submit(job.action, **job.kwargs))
                    job.future.add_done_callback(self.job_callback)
                else:
                    target_job_tracker.return_to_queue(job)

        if len(self.errors) > 0 and self.soft_failure:
            Logger.instance().error(
                f"\nRefresh failed in {round(time()-start_time, 2)}s with {len(self.errors)} query error(s)."
            )
        elif len(self.errors) > 0 and not self.soft_failure:
            Logger.instance().error(
                f"\nRun failed in {round(time()-start_time, 2)}s with {len(self.errors)} query error(s)"
            )
            exit(1)
        else:
            Logger.instance().info(f"\nRun finished in {round(time()-start_time, 2)}s")

    def update_job_queue(self, target_job_tracker: TargetJobTracker) -> bool:
        all_dependencies_completed = True
        for job in self.jobs:
            job_item_children = ParentModel.all_descendants(
                dag=self.dag, from_node=job.item
            )
            dependencies = list(
                filter(
                    lambda j: job != j and job_item_children.has_node(j.item),
                    self.jobs,
                )
            )
            incomplete_dependencies = list(
                filter(
                    lambda d: not target_job_tracker.is_job_name_done(d.name),
                    dependencies,
                )
            )
            dependencies_completed = len(incomplete_dependencies) == 0
            if not dependencies_completed:
                all_dependencies_completed = False

            if dependencies_completed and not target_job_tracker.is_job_name_enqueued(
                job.name
            ):
                if not job.output_changed and self.run_only_changed:
                    job.future = CachedFuture()
                target_job_tracker.track_job(job)

        return all_dependencies_completed and target_job_tracker.empty()

    def job_callback(self, future: Future):
        job_result: JobResult = future.result(timeout=1)
        if job_result.success:
            Logger.instance().success(str(job_result.message))
        else:
            Logger.instance().error(str(job_result.message))
            self.errors.append(str(job_result.message))

    def _all_jobs(self) -> List[Job]:
        jobs = []
        jobs = jobs + run_trace_jobs(
            dag=self.dag,
            output_dir=self.output_dir,
            project=self.project,
            name_filter=self.name_filter,
        )
        jobs = jobs + csv_script_jobs(
            dag=self.dag,
            output_dir=self.output_dir,
            project=self.project,
            name_filter=self.name_filter,
        )
        jobs = jobs + run_local_merge_jobs(
            dag=self.dag,
            output_dir=self.output_dir,
            project=self.project,
            name_filter=self.name_filter,
        )
        return jobs



import click
from visivo.commands.options import output_dir, working_dir, target, alert


@click.command()
@target
@working_dir
@output_dir
@alert
def test(output_dir, working_dir, target, alert):
    """
    Enables testing trace values to ensure that the charts that are being produced have the characteristics that you would expect.
    """
    from visivo.logging.logger import Logger

    Logger.instance().debug("Testing")

    from visivo.commands.test_phase import test_phase

    test_phase(
        default_target=target,
        output_dir=output_dir,
        working_dir=working_dir,
        alert_names=alert,
    )
    Logger.instance().success("Done")



import yaml
import json
from collections import OrderedDict


class YamlOrderedDict(OrderedDict):
    """
    An OrderedDict that was loaded from a yaml file, and is annotated
    with file/line info for reporting about errors in the source file
    """

    def _annotate(self, node):
        self._key_locs = {}
        self._value_locs = {}
        nodeiter = node.value.__iter__()
        for key in self:
            subnode = nodeiter.__next__()
            self._key_locs[key] = (
                subnode[0].start_mark.name + ":" + str(subnode[0].start_mark.line + 1)
            )
            self._value_locs[key] = (
                subnode[1].start_mark.name + ":" + str(subnode[1].start_mark.line + 1)
            )

    def key_loc(self, key):
        try:
            return self._key_locs[key]
        except (AttributeError, KeyError):
            return ""

    def value_loc(self, key):
        try:
            return self._value_locs[key]
        except (AttributeError, KeyError):
            return ""

    def __str__(self):
        return json.dumps(self)


def setup_yaml_ordered_dict():
    # Use YamlOrderedDict objects for yaml maps instead of normal dict
    yaml.add_representer(
        OrderedDict, lambda dumper, data: dumper.represent_dict(data.items())
    )
    yaml.add_representer(
        YamlOrderedDict, lambda dumper, data: dumper.represent_dict(data.items())
    )

    def _load_YamlOrderedDict(loader, node):
        rv = YamlOrderedDict(loader.construct_pairs(node))
        rv._annotate(node)
        return rv

    yaml.add_constructor(
        yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG,
        _load_YamlOrderedDict,
        yaml.SafeLoader,
    )



from typing import Any, List, Optional

from visivo.models.base.selector_model import SelectorModel
from visivo.models.selector import Selector
from visivo.models.table_column_definition import TableColumnDefinition
from .trace import Trace
from pydantic import Field
from .base.named_model import NamedModel
from .base.parent_model import ParentModel
from .base.base_model import REF_REGEX, generate_ref_field
from pydantic import model_validator


class Table(SelectorModel, NamedModel, ParentModel):
    """
    Tables enable you to quickly represent trace data in a tabular format.

    Since tables sit on top of trace data, the steps to create a table from scratch are as follows:

    1. Create a model.
    1. Create a trace with columns or props that references your model.
    1. Create a table that references the trace. Within the table.columns block you will need to explicitly state the trace columns and header names that you want to include.

    ### Example
    ``` yaml
    models:
      - name: table-model
        sql: |
            select
                project_name,
                project_created_at,
                cli_version,
                stage_name,
                account_name,
                stage_archived
            FROM visivo_project
    traces:
      - name: pre-table-trace
        model: ref(table-model)
        columns:
            project_name: project_name
            project_created_at: project_created_at::varchar
            cli_version: cli_version
            stage_name: stage_name
            account_name: account_name
            stage_archived: stage_archived::varchar
        props:
            type: scatter
            x: column(project_created_at)
            y: column(project_name)
    tables:
      - name: latest-projects-table
        traces:
          - ref(pre-table-trace)
        column_defs:
          - trace_name: pre-table-trace
            columns:
            - header: "Project Name"
              key: columns.project_name
            - header: "Project Created At"
              key: columns.project_created_at
            - header: "Project Json"
              key: columns.project_json
            - header: "CLI Version"
              key: columns.cli_version
            - header: "Stage Name"
              key: columns.stage_name
            - header: "Account Name"
              key: columns.account_name
            - header: "Account Name"
              key: columns.stage_archived
    ```
    Tables are built on the [material react table framework](https://www.material-react-table.com/).
    """

    traces: List[generate_ref_field(Trace)] = Field(
        [],
        description="A ref() to a trace or trace defined in line.  Data for the table will come from the trace.",
    )

    column_defs: Optional[List[TableColumnDefinition]] = Field(
        None,
        description="A list of column definitions. These definitions define the columns for a given trace included in this table.",
    )

    def child_items(self):
        return self.traces + [self.selector]

    @property
    def trace_objs(self) -> List[Trace]:
        return list(filter(Trace.is_obj, self.traces))

    @property
    def trace_refs(self) -> List[str]:
        return list(filter(Trace.is_ref, self.traces))

    @model_validator(mode="before")
    @classmethod
    def validate_column_defs(cls, data: any):
        traces, column_defs = (data.get("traces"), data.get("column_defs"))

        if not column_defs:
            return data

        column_defs_trace_names = list(map(lambda cd: cd["trace_name"], column_defs))
        traces_trace_names = list(map(lambda t: NamedModel.get_name(t), traces))
        for column_defs_trace_name in column_defs_trace_names:
            if not column_defs_trace_name in traces_trace_names:
                raise ValueError(
                    f"Column def trace name '{column_defs_trace_name}' is not present in trace list on table."
                )
        return data

    @model_validator(mode="before")
    @classmethod
    def ensure_selector(cls, data: Any) -> Any:
        selector = data.get("selector")
        if selector is None:
            name = data.get("name")
            data["selector"] = {"name": f"{name} Selector", "type": "single"}

        return data



# supports more generic connections than the snowflake specific connector
from visivo.models.project import Project
from visivo.models.trace import Trace
from visivo.models.test_run import TestRun, TestFailure, TestSuccess
from visivo.models.alert import Alert
from typing import List
import json
import click
from datetime import datetime
import warnings


warnings.filterwarnings("ignore")


class Runner:
    def __init__(
        self,
        traces: List[Trace],
        project: Project,
        output_dir: str,
        default_target: str = None,
        alerts: List[Alert] = [],
    ):
        self.project = project
        self.traces = traces
        self.default_target = default_target
        self.output_dir = output_dir
        self.alerts = alerts

    def run(self):
        import numpy
        from assertpy import assert_that
        import os

        test_run = TestRun()
        for trace in self.traces:
            if not trace.tests:
                continue
            for idx, test in enumerate(trace.tests):
                data_file = f"{self.output_dir}/{trace.name}/data.json"
                if not os.path.exists(data_file):
                    raise click.ClickException(
                        f"The trace '{trace.name}' doesn't have a data file present, please run 'visivo run'."
                    )
                with open(data_file) as f:
                    trace_data = json.load(f)
                    logic = test.logic
                    try:
                        if not any(k in logic for k in trace_data.keys()):
                            raise click.ClickException(
                                f"The test does not reference a valid data cohort.  Available cohorts: {', '.join(trace_data.keys())}"
                            )
                        for cohort in trace_data.keys():
                            if cohort in logic:
                                logic = logic.replace(cohort, f"trace_data['{cohort}']")
                                cohort_keys = list(trace_data[cohort].keys())
                                cohort_keys.sort(key=len, reverse=True)
                                for key in cohort_keys:
                                    logic = logic.replace(key, f"['{key}']")
                        logic = logic.replace("].[", "][")
                        eval(f"{logic}")
                        click.echo(click.style(".", fg="green"), nl=False)
                        success = TestSuccess(test_id=f"{trace.name}.test[{idx}]")
                        test_run.add_success(success=success)
                    except Exception as e:
                        failure = TestFailure(
                            test_id=f"{trace.name}.test[{idx}]", message=str(e)
                        )
                        click.echo(click.style("F", fg="red"), nl=False)
                        test_run.add_failure(failure=failure)
        test_run.finished_at = datetime.now()
        click.echo("")
        click.echo(test_run.summary())

        for alert in self.alerts:
            alert.alert(test_run=test_run)

        return test_run



from typing import Optional

from visivo.models.base.base_model import generate_ref_field
from pydantic import Field
from visivo.models.base.parent_model import ParentModel
from visivo.models.models.model import Model
from visivo.models.targets.fields import TargetRefField
from visivo.models.targets.target import DefaultTarget, Target


class SqlModel(Model, ParentModel):
    """
    SQL Models are queries that return base data from a SQL target. These data are then used in Traces

    !!! example {% raw %}

        === "Select All"

            ``` yaml
            models:
              - name: sql_model
                target: ref(sql_target)
                sql: select * from table_in_target
            ```
    {% endraw %}
    """

    sql: str = Field(
        None,
        description="The sql used to generate your base data",
    )

    target: Optional[TargetRefField] = Field(
        None,
        description="A target object defined inline or a ref() to a chart. Override the defaults.target_name",
    )

    def child_items(self):
        if self.target:
            return [self.target]
        else:
            return [DefaultTarget()]



from visivo.models.base.parent_model import ParentModel
from visivo.models.models.csv_script_model import CsvScriptModel
from visivo.models.project import Project
from visivo.query.jobs.job import (
    Job,
    JobResult,
    format_message_failure,
    format_message_success,
)
from time import time


def action(csv_script_model: CsvScriptModel, output_dir):
    try:
        start_time = time()
        csv_script_model.insert_csv_to_sqlite(output_dir=output_dir)
        success_message = format_message_success(
            details=f"Updated data for model \033[4m{csv_script_model.name}\033[0m",
            start_time=start_time,
            full_path=csv_script_model.get_sqlite_target(
                output_dir=output_dir
            ).database,
        )
        return JobResult(success=True, message=success_message)
    except Exception as e:
        failure_message = format_message_failure(
            details=f"Failed query for model \033[4m{csv_script_model.name}\033[0m",
            start_time=start_time,
            full_path=csv_script_model.get_sqlite_target(
                output_dir=output_dir
            ).database,
            error_msg=str(repr(e)),
        )
        return JobResult(success=False, message=failure_message)


def jobs(dag, output_dir: str, project: Project, name_filter: str):
    csv_script_models = ParentModel.all_descendants_of_type(
        type=CsvScriptModel, dag=dag, from_node=project
    )

    if name_filter:
        included_nodes = project.nodes_including_named_node_in_graph(name=name_filter)
    else:
        included_nodes = project.descendants()
    csv_script_models = set(csv_script_models).intersection(included_nodes)

    jobs = []
    for csv_script_model in csv_script_models:
        jobs.append(
            Job(
                item=csv_script_model,
                target=csv_script_model.get_sqlite_target(output_dir),
                action=action,
                csv_script_model=csv_script_model,
                output_dir=output_dir,
            )
        )
    return jobs



import click
from visivo.commands.options import (
    output_dir,
    working_dir,
    target,
    name_filter,
    threads,
)


@click.command()
@target
@working_dir
@output_dir
@name_filter
@threads
def run(output_dir, working_dir, target, name_filter, threads):
    """
    Compiles the project and then runs the trace queries to fetch data to populate in the traces. Writes all data to the target directory.
    """

    from visivo.logging.logger import Logger

    Logger.instance().debug("Running")

    from visivo.commands.run_phase import run_phase

    run_phase(
        default_target=target,
        output_dir=output_dir,
        working_dir=working_dir,
        name_filter=name_filter,
        threads=threads,
    )
    Logger.instance().success("Done")



from enum import Enum
from pydantic import ConfigDict, Field, PrivateAttr, model_serializer

from visivo.models.base.base_model import BaseModel
from visivo.models.base.named_model import NamedModel


class SelectorType(str, Enum):
    single = "single"
    multiple = "multiple"


class Selector(NamedModel, BaseModel):
    """
    Selectors can be used to add interactivity between charts and tables.

    Below is how you would link two charts to show the same selected data.

    ### Example

    ``` yaml
    charts:
        - name: Chart One
          selector:
              name: Common Selector
              type: single
              ...
        - name: Chart Two
          selector: ref(Common Selector)
          ...
    ```

    """
    model_config = ConfigDict(extra='ignore')  

    name: str = Field(description="The name of the selector")
    type: SelectorType = Field(
        SelectorType.multiple, description="Single or multiple selector"
    )
    _parent_name: str = PrivateAttr()

    def set_parent_name(self, value: str):
        self._parent_name = value

    @model_serializer()
    def serialize_model(self):
        return {"name": self.name, "type": self.type, "parent_name": self._parent_name}



import click


@click.command()
def init():
    """
    Enables a quick set up by writing your target & api credentials to an env file.
    """
    from visivo.logging.logger import Logger

    Logger.instance().debug("Initializing")

    from visivo.commands.init_phase import init_phase

    init_phase()
    Logger.instance().success("Done")



import click
import os
from pathlib import Path
from visivo.parsers.file_names import PROFILE_FILE_NAME
from visivo.utils import load_yaml_file


def get_profile_token(profile_file):
    profile_token = os.getenv("VISIVO_TOKEN")
    if profile_token:
        return profile_token

    profile = None
    if profile_file:
        profile = load_yaml_file(profile_file)

    if not profile or "token" not in profile:
        raise click.ClickException(
            f"{PROFILE_FILE_NAME} not present or token not present in {PROFILE_FILE_NAME}"
        )
    return profile["token"]


def get_profile_file(home_directory=os.path.expanduser("~")):
    return Path(f"{home_directory}/.visivo/{PROFILE_FILE_NAME}")


def create_file_database(url, output_dir: str):
    from sqlalchemy import create_engine, MetaData, Table, Integer, Column, insert

    os.makedirs(output_dir, exist_ok=True)
    engine = create_engine(url, echo=True)
    metadata_obj = MetaData()
    table = Table(
        "test_table",
        metadata_obj,
        Column("x", Integer),
        Column("y", Integer),
    )
    second_table = Table(
        "second_test_table",
        metadata_obj,
        Column("x", Integer),
        Column("y", Integer),
    )
    metadata_obj.create_all(engine)
    for v in [[1, 1], [2, 1], [3, 2], [4, 3], [5, 5], [6, 8]]:
        with engine.connect() as connection:
            connection.execute(insert(table).values(x=v[0], y=v[1]))
            connection.execute(insert(second_table).values(x=v[0], y=v[1] * 2))
            connection.commit()



from typing import List, Optional
from pydantic import Field
from pathlib import Path

from visivo.models.base.base_model import BaseModel


class TableColumn(BaseModel):
    header: Optional[str] = Field(
        None,
        description="The display name of the column.  Defaults to readable key name.",
    )
    key: str = Field(
        None,
        description="The key to the data that will pull back the value.",
    )


class TableColumnDefinition(BaseModel):
    """
    Table columns allow you to display a subset of the available trace data.
    """

    trace_name: str = Field(
        None,
        description="The name of the trace that the column defs apply to.",
    )
    columns: List[TableColumn] = Field(
        description="A list of column definitions that contain `header` and `key`. `header` is the title of the column in the table. `key` is the path to the array property you want to include.  For example 'props.x' or 'columns.x_data'.",
    )

    @property
    def file(self):
        return Path(self.path)



import click
import requests
import json
from visivo.commands.utils import get_profile_file, get_profile_token
from visivo.discovery.discover import Discover
from visivo.logging.logger import Logger
from visivo.parsers.serializer import Serializer
from visivo.parsers.parser_factory import ParserFactory


def deploy_phase(working_dir, user_dir, output_dir, stage, host):
    """
    Sends the current version of your project, traces & data to app.visivo.io where it can be viewed by other users on your account. You must specify a stage when deploying a project. The stage allows multiple versions of your project to exist remotely. This is very useful for setting up different dev, CI and production environments.
    """
    profile_token = get_profile_token(get_profile_file(home_directory=user_dir))

    discover = Discover(working_directory=working_dir, home_directory=user_dir)
    parser = ParserFactory().build(
        project_file=discover.project_file, files=discover.files
    )
    project = parser.parse()
    serializer = Serializer(project=project)
    project_json = json.loads(
        serializer.dereference().model_dump_json(exclude_none=True)
    )

    body = {
        "project_json": project_json,
        "name": project_json["name"],
        "cli_version": project_json["cli_version"],
        "stage": stage,
    }
    json_headers = {
        "content-type": "application/json",
        "Authorization": f"Api-Key {profile_token}",
    }
    form_headers = {
        "Authorization": f"Api-Key {profile_token}",
    }

    url = f"{host}/api/projects/"
    response = requests.post(url, data=json.dumps(body), headers=json_headers)
    if response.status_code == 401:
        raise click.ClickException(f"Token not authorized for host: {host}")
    if response.status_code == 404:
        raise click.ClickException(f"404 error raised. Does your user have an account?")
    if response.status_code == 201:
        Logger.instance().debug("Project uploaded")
        project_data = response.json()
        project_id = project_data["id"]
        project_url = project_data["url"]

        for trace in project.trace_objs:
            url = f"{host}/api/files/"

            data_file = f"{output_dir}/{trace.name}/data.json"
            files = {"file": open(data_file, "rb")}
            response = requests.post(url, files=files, data={}, headers=form_headers)
            if response.status_code != 201:
                raise click.ClickException(f"Trace '{trace.name}' data not uploaded")
            Logger.instance().debug(f"Trace '{trace.name}' data uploaded")
            url = f"{host}/api/traces/"
            body = {
                "name": trace.name,
                "project_id": project_id,
                "data_file_id": response.json()["id"],
            }
            response = requests.post(url, data=json.dumps(body), headers=json_headers)
            if response.status_code != 201:
                Logger.instance().debug(response.json())
                raise click.ClickException(f"Trace '{trace.name}' not created")
            Logger.instance().debug(f"Trace '{trace.name}' created")
        return project_url
    else:
        raise click.ClickException(f"There was an unexpected error: {response.content}")



from typing import List
from visivo.models.models.model import Model, TableModelName
from pydantic import Field
from visivo.models.targets.sqlite_target import SqliteTarget
import io
import click


class CsvScriptModel(Model):
    """
    CSV Script Models are a type of model that executes a command with a given set of args.
    This command needs to return a well formatted :fontawesome-solid-file-csv: with a header row to stdout.

    Visivo will be able to access the generate file as a model by storing a sqlite file in the target directory.

    !!! example {% raw %}

        === "Echo"

            Echoing all of your data is probably not a very practical example, but it does nicely demonstrate how the feature works!
            ``` yaml
            models:
              - name: csv
                table_name: csv
                args:
                    - echo
                    - |
                      x,y
                      1,9
                      2,1
                      3,2
                      4,3
                      5,5
                      6,8
            ```

        === "Python Script"

            In this example we'll use python to generate a csv of processes running on your machine and make that csv available to Visivo as
            a model for analysis.
            ``` python title="created_processes_csv.py"
            import subprocess
            import csv
            import sys

            # Define the CSV file to write
            csv_file = "target/processes.csv"

            # Execute the 'ps aux' command
            result = subprocess.run(["ps", "aux"], stdout=subprocess.PIPE, text=True)

            # Split the output into lines
            lines = result.stdout.strip().split("/n")

            # Write CSV to stdout
            writer = csv.writer(sys.stdout)
            writer.writerow(
                ["USER","PID","%CPU","%MEM","VSZ","RSS","TTY","STAT","START","TIME","COMMAND"]
            )  # Header

            for line in lines[1:]:  # Skip the header line from the ps output
                row = line.split(None, 10)  # Split on whitespace, but only for the first 10 columns
                writer.writerow(row)
            ```
            With your script ready to go, all you have to do is convert `python create_processes_csv.py` into the args list format in a model.
            ``` yaml
            models:
              - name: processes
                table_name: processes
                args:
                  - python
                  - create_processes_csv.py
            ```{% endraw %}

        === "CSV File"

            One of the best use cases for this type of model is to store a static csv in your project and cat it into a model.
            This great because it's simple and allows you to version control your csv data.
            ``` csv title="file.csv"
            columns,go,up,here
            1,text,more text,6
            2,stuff,more stuff,7
            ```
            Then just `cat` the csv file in a model.
            ``` yaml
            models:
              - name: file_model
                table_name: file_model
                args:
                  - cat
                  - file.csv
            ```

    The args are python subprocess list args and you can read their source [documentation here](https://docs.python.org/3/library/subprocess.html#subprocess.CompletedProcess.args).
    """

    table_name: TableModelName = Field(
        "model", description="The name to give the resulting models table"
    )

    args: List[str] = Field(
        description="An array of the variables that build your command to run."
    )

    @property
    def sql(self):
        return f"select * from {self.table_name}"

    def get_sqlite_target(self, output_dir) -> SqliteTarget:
        return SqliteTarget(
            name=f"model_{self.name}_generated_target",
            database=f"{output_dir}/{self.name}.sqlite",
            type="sqlite",
        )

    def insert_csv_to_sqlite(self, output_dir):
        import pandas
        import subprocess

        process = subprocess.Popen(self.args, stdout=subprocess.PIPE)
        engine = self.get_sqlite_target(output_dir).get_engine()
        try:
            csv = io.StringIO(process.stdout.read().decode())
            data_frame = pandas.read_csv(csv)
        except:
            raise click.ClickException(
                f"Error parsing csv output of {self.name} model's command. Verify command's output and try again."
            )
        data_frame.to_sql(self.table_name, engine, if_exists="replace", index=False)



from deepmerge import always_merger
from typing import List
from pathlib import Path
from pydantic import ValidationError
from visivo.parsers.line_validation_error import LineValidationError
from visivo.parsers.yaml_ordered_dict import setup_yaml_ordered_dict
from visivo.utils import load_yaml_file
from ..models.project import Project


class CoreParser:
    def __init__(self, project_file: Path, files: List[Path]):
        self.files = files
        self.project_file = project_file
        setup_yaml_ordered_dict()

    def parse(self) -> Project:
        return self.__build_project()

    def project_file_data(self):
        return load_yaml_file(self.project_file)

    def __build_project(self):
        data = self.__merged_project_data()
        try:
            project = Project(**data)
        except ValidationError as validation_error:
            raise LineValidationError(
                validation_error=validation_error, files=self.files
            )
        return project

    def __merged_project_data(self):
        project_data = self.project_file_data()

        data_files = []
        for file in self.files:
            if file == self.project_file:
                continue
            data_files.append(load_yaml_file(file))

        return self.__merge_data_into_project(
            project_data=project_data, data_files=data_files
        )

    def __merge_data_into_project(self, project_data: dict, data_files: List[dict]):
        keys_to_merge = [
            "alerts",
            "targets",
            "models",
            "traces",
            "tables",
            "charts",
            "dashboards",
        ]
        for data_file in data_files:
            for key_to_merge in keys_to_merge:
                if key_to_merge in data_file:
                    base_merge = []
                    if key_to_merge in project_data:
                        base_merge = project_data[key_to_merge]
                    project_data[key_to_merge] = always_merger.merge(
                        base_merge, data_file[key_to_merge]
                    )
        return project_data



import sys
import click
from visivo.logging.logger import Logger
from visivo.testing.runner import Runner
from visivo.commands.compile_phase import compile_phase


def test_phase(
    output_dir: str, default_target: str, working_dir: str, alert_names: str
):
    project = compile_phase(
        default_target=default_target, working_dir=working_dir, output_dir=output_dir
    )
    Logger.instance().debug("Testing project")
    alerts = list(map(lambda an: project.find_alert(name=an), alert_names))
    alerts = list(filter(None, alerts))
    test_runner = Runner(
        traces=project.trace_objs,
        project=project,
        output_dir=output_dir,
        alerts=alerts,
        default_target=default_target,
    )
    if not test_runner.run().success:
        sys.exit(1)


test_phase.__test__ = False




from pydantic import Field, constr
from visivo.models.base.base_model import INDEXED_STATEMENT_REGEX, STATEMENT_REGEX
from visivo.models.trace_props.trace_props import  TraceProps, TracePropsAttribute
from typing import List, Literal, Optional 


class Line2(TracePropsAttribute):
	color: Optional[str]= Field(
		None,
		description=""" color<br>Sets the line color. """
	)
	dash: Optional[str]= Field(
		None,
		description=""" string<br>Sets the dash style of lines. Set to a dash type string ("solid", "dot", "dash", "longdash", "dashdot", or "longdashdot") or a dash length list in px (eg "5px,10px,2px,2px"). """
	)
	width: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number greater than or equal to 0<br>Sets the line width (in px). """
	)
class WaterfallConnector(TracePropsAttribute):
	line: Optional[Line2]= Field(
		None,
		description=""" object containing one or more of the keys listed below.<br> """
	)
	mode: Optional[str]= Field(
		None,
		description=""" enumerated , one of ( "spanning" | "between" )<br>Sets the shape of connector lines. """
	)
	visible: Optional[bool | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" boolean<br>Determines if connector lines are drawn.  """
	)
class Line24(TracePropsAttribute):
	color: Optional[str]= Field(
		None,
		description=""" color<br>Sets the line color of all decreasing values. """
	)
	width: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number greater than or equal to 0<br>Sets the line width of all decreasing values. """
	)
class WaterfallDecreasingMarker(TracePropsAttribute):
	color: Optional[str]= Field(
		None,
		description=""" color<br>Sets the marker color of all decreasing values. """
	)
	line: Optional[Line24]= Field(
		None,
		description=""" object containing one or more of the keys listed below.<br> """
	)
class WaterfallDecreasing(TracePropsAttribute):
	marker: Optional[WaterfallDecreasingMarker]= Field(
		None,
		description=""" object containing one or more of the keys listed below.<br> """
	)
class FontInsidetextfontTextfontOutsidetextfont1(TracePropsAttribute):
	color: Optional[str | List[str]]= Field(
		None,
		description=""" color or array of colors<br> """
	)
	family: Optional[str | List[str]]= Field(
		None,
		description=""" string or array of strings<br>HTML font family - the typeface that will be applied by the web browser. The web browser will only be able to apply a font if it is available on the system which it operates. Provide multiple font families, separated by commas, to indicate the preference in which to apply fonts if they aren't available on the system. The Chart Studio Cloud (at https://chart-studio.plotly.com or on-premise) generates images on a server, where only a select number of fonts are installed and supported. These include "Arial", "Balto", "Courier New", "Droid Sans",, "Droid Serif", "Droid Sans Mono", "Gravitas One", "Old Standard TT", "Open Sans", "Overpass", "PT Sans Narrow", "Raleway", "Times New Roman". """
	)
	size: Optional[constr(pattern=INDEXED_STATEMENT_REGEX) | constr(pattern=STATEMENT_REGEX) | float | List[float]]= Field(
		None,
		description=""" number or array of numbers greater than or equal to 1<br> """
	)
class WaterfallHoverlabel(TracePropsAttribute):
	align: Optional[str | List[str] ]= Field(
		None,
		description=""" enumerated or array of enumerateds , one of ( "left" | "right" | "auto" )<br>Sets the horizontal alignment of the text content within hover label box. Has an effect only if the hover label text spans more two or more lines """
	)
	bgcolor: Optional[str | List[str]]= Field(
		None,
		description=""" color or array of colors<br>Sets the background color of the hover labels for this trace """
	)
	bordercolor: Optional[str | List[str]]= Field(
		None,
		description=""" color or array of colors<br>Sets the border color of the hover labels for this trace. """
	)
	font: Optional[FontInsidetextfontTextfontOutsidetextfont1]= Field(
		None,
		description=""" object containing one or more of the keys listed below.<br>Sets the font used in hover labels. """
	)
	namelength: Optional[int | constr(pattern=INDEXED_STATEMENT_REGEX) | constr(pattern=STATEMENT_REGEX) | List[int]]= Field(
		None,
		description=""" integer or array of integers greater than or equal to -1<br>Sets the default length (in number of characters) of the trace name in the hover labels for all traces. -1 shows the whole name regardless of length. 0-3 shows the first 0-3 characters, and an integer >3 will show the whole name if it is less than that many characters, but if it is longer, will truncate to `namelength - 3` characters and add an ellipsis. """
	)
class Line27(TracePropsAttribute):
	color: Optional[str]= Field(
		None,
		description=""" color<br>Sets the line color of all increasing values. """
	)
	width: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number greater than or equal to 0<br>Sets the line width of all increasing values. """
	)
class WaterfallIncreasingMarker(TracePropsAttribute):
	color: Optional[str]= Field(
		None,
		description=""" color<br>Sets the marker color of all increasing values. """
	)
	line: Optional[Line27]= Field(
		None,
		description=""" object containing one or more of the keys listed below.<br> """
	)
class WaterfallIncreasing(TracePropsAttribute):
	marker: Optional[WaterfallIncreasingMarker]= Field(
		None,
		description=""" object containing one or more of the keys listed below.<br> """
	)
class FontInsidetextfontTextfontOutsidetextfont1(TracePropsAttribute):
	color: Optional[str | List[str]]= Field(
		None,
		description=""" color or array of colors<br> """
	)
	family: Optional[str | List[str]]= Field(
		None,
		description=""" string or array of strings<br>HTML font family - the typeface that will be applied by the web browser. The web browser will only be able to apply a font if it is available on the system which it operates. Provide multiple font families, separated by commas, to indicate the preference in which to apply fonts if they aren't available on the system. The Chart Studio Cloud (at https://chart-studio.plotly.com or on-premise) generates images on a server, where only a select number of fonts are installed and supported. These include "Arial", "Balto", "Courier New", "Droid Sans",, "Droid Serif", "Droid Sans Mono", "Gravitas One", "Old Standard TT", "Open Sans", "Overpass", "PT Sans Narrow", "Raleway", "Times New Roman". """
	)
	size: Optional[constr(pattern=INDEXED_STATEMENT_REGEX) | constr(pattern=STATEMENT_REGEX) | float | List[float]]= Field(
		None,
		description=""" number or array of numbers greater than or equal to 1<br> """
	)
class TextfontLabelfontTickfontInsidetextfontRangefontOutsidetextfontFontGrouptitlefont1(TracePropsAttribute):
	color: Optional[str]= Field(
		None,
		description=""" color<br> """
	)
	family: Optional[str]= Field(
		None,
		description=""" string<br>HTML font family - the typeface that will be applied by the web browser. The web browser will only be able to apply a font if it is available on the system which it operates. Provide multiple font families, separated by commas, to indicate the preference in which to apply fonts if they aren't available on the system. The Chart Studio Cloud (at https://chart-studio.plotly.com or on-premise) generates images on a server, where only a select number of fonts are installed and supported. These include "Arial", "Balto", "Courier New", "Droid Sans",, "Droid Serif", "Droid Sans Mono", "Gravitas One", "Old Standard TT", "Open Sans", "Overpass", "PT Sans Narrow", "Raleway", "Times New Roman". """
	)
	size: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number greater than or equal to 1<br> """
	)
class WaterfallLegendgrouptitle(TracePropsAttribute):
	font: Optional[TextfontLabelfontTickfontInsidetextfontRangefontOutsidetextfontFontGrouptitlefont1]= Field(
		None,
		description=""" object containing one or more of the keys listed below.<br>Sets this legend group's title font. """
	)
	text: Optional[str]= Field(
		None,
		description=""" string<br>Sets the title of the legend group. """
	)
class FontInsidetextfontTextfontOutsidetextfont1(TracePropsAttribute):
	color: Optional[str | List[str]]= Field(
		None,
		description=""" color or array of colors<br> """
	)
	family: Optional[str | List[str]]= Field(
		None,
		description=""" string or array of strings<br>HTML font family - the typeface that will be applied by the web browser. The web browser will only be able to apply a font if it is available on the system which it operates. Provide multiple font families, separated by commas, to indicate the preference in which to apply fonts if they aren't available on the system. The Chart Studio Cloud (at https://chart-studio.plotly.com or on-premise) generates images on a server, where only a select number of fonts are installed and supported. These include "Arial", "Balto", "Courier New", "Droid Sans",, "Droid Serif", "Droid Sans Mono", "Gravitas One", "Old Standard TT", "Open Sans", "Overpass", "PT Sans Narrow", "Raleway", "Times New Roman". """
	)
	size: Optional[constr(pattern=INDEXED_STATEMENT_REGEX) | constr(pattern=STATEMENT_REGEX) | float | List[float]]= Field(
		None,
		description=""" number or array of numbers greater than or equal to 1<br> """
	)
class FontInsidetextfontTextfontOutsidetextfont1(TracePropsAttribute):
	color: Optional[str | List[str]]= Field(
		None,
		description=""" color or array of colors<br> """
	)
	family: Optional[str | List[str]]= Field(
		None,
		description=""" string or array of strings<br>HTML font family - the typeface that will be applied by the web browser. The web browser will only be able to apply a font if it is available on the system which it operates. Provide multiple font families, separated by commas, to indicate the preference in which to apply fonts if they aren't available on the system. The Chart Studio Cloud (at https://chart-studio.plotly.com or on-premise) generates images on a server, where only a select number of fonts are installed and supported. These include "Arial", "Balto", "Courier New", "Droid Sans",, "Droid Serif", "Droid Sans Mono", "Gravitas One", "Old Standard TT", "Open Sans", "Overpass", "PT Sans Narrow", "Raleway", "Times New Roman". """
	)
	size: Optional[constr(pattern=INDEXED_STATEMENT_REGEX) | constr(pattern=STATEMENT_REGEX) | float | List[float]]= Field(
		None,
		description=""" number or array of numbers greater than or equal to 1<br> """
	)
class Line26(TracePropsAttribute):
	color: Optional[str]= Field(
		None,
		description=""" color<br>Sets the line color of all intermediate sums and total values. """
	)
	width: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number greater than or equal to 0<br>Sets the line width of all intermediate sums and total values. """
	)
class WaterfallTotalsMarker(TracePropsAttribute):
	color: Optional[str]= Field(
		None,
		description=""" color<br>Sets the marker color of all intermediate sums and total values. """
	)
	line: Optional[Line26]= Field(
		None,
		description=""" object containing one or more of the keys listed below.<br> """
	)
class WaterfallTotals(TracePropsAttribute):
	marker: Optional[WaterfallTotalsMarker]= Field(
		None,
		description=""" object containing one or more of the keys listed below.<br> """
	)
class Waterfall(TraceProps):
	alignmentgroup: Optional[str]= Field(
		None,
		description=""" string<br>Set several traces linked to the same position axis or matching axes to the same alignmentgroup. This controls whether bars compute their positional range dependently or independently. """
	)
	base: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number<br>Sets where the bar base is drawn (in position axis units). """
	)
	cliponaxis: Optional[bool | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" boolean<br>Determines whether the text nodes are clipped about the subplot axes. To show the text nodes above axis lines and tick labels, make sure to set `xaxis.layer` and `yaxis.layer` to "below traces". """
	)
	connector: Optional[WaterfallConnector]= Field(
		None,
		description=""" object containing one or more of the keys listed below.<br> """
	)
	constraintext: Optional[str]= Field(
		None,
		description=""" enumerated , one of ( "inside" | "outside" | "both" | "none" )<br>Constrain the size of text inside or outside a bar to be no larger than the bar itself. """
	)
	customdata: Optional[constr(pattern=STATEMENT_REGEX) | List]= Field(
		None,
		description=""" data array<br>Assigns extra data each datum. This may be useful when listening to hover, click and selection events. Note that, "scatter" traces also appends customdata items in the markers DOM elements """
	)
	decreasing: Optional[WaterfallDecreasing]= Field(
		None,
		description=""" object containing one or more of the keys listed below.<br> """
	)
	dx: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number<br>Sets the x coordinate step. See `x0` for more info. """
	)
	dy: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number<br>Sets the y coordinate step. See `y0` for more info. """
	)
	hoverinfo: Optional[str]= Field(
		None,
		description=""" flaglist string. any combination of "name", "x", "y", "text", "initial", "delta", "final" joined with a "+" or "all" or "none" or "skip".<br>Determines which trace information appear on hover. If `none` or `skip` are set, no information is displayed upon hovering. But, if `none` is set, click and hover events are still fired. """
	)
	hoverlabel: Optional[WaterfallHoverlabel]= Field(
		None,
		description=""" object containing one or more of the keys listed below.<br> """
	)
	hovertemplate: Optional[str | List[str]]= Field(
		None,
		description=""" string or array of strings<br>Template string used for rendering the information that appear on hover box. Note that this will override `hoverinfo`. Variables are inserted using %{variable}, for example "y: %{y}" as well as %{xother}, {%_xother}, {%_xother_}, {%xother_}. When showing info for several points, "xother" will be added to those with different x positions from the first point. An underscore before or after "(x|y)other" will add a space on that side, only when this field is shown. Numbers are formatted using d3-format's syntax %{variable:d3-format}, for example "Price: %{y:$.2f}". https://github.com/d3/d3-format/tree/v1.4.5#d3-format for details on the formatting syntax. Dates are formatted using d3-time-format's syntax %{variable|d3-time-format}, for example "Day: %{2019-01-01|%A}". https://github.com/d3/d3-time-format/tree/v2.2.3#locale_format for details on the date formatting syntax. The variables available in `hovertemplate` are the ones emitted as event data described at this link https://plotly.com/javascript/plotlyjs-events/#event-data. Additionally, every attributes that can be specified per-point (the ones that are `arrayOk: true`) are available. variables `initial`, `delta` and `final`. Anything contained in tag `<extra>` is displayed in the secondary box, for example "<extra>{fullData.name}</extra>". To hide the secondary box completely, use an empty tag `<extra></extra>`. """
	)
	hovertext: Optional[str | List[str]]= Field(
		None,
		description=""" string or array of strings<br>Sets hover text elements associated with each (x,y) pair. If a single string, the same string appears over all the data points. If an array of string, the items are mapped in order to the this trace's (x,y) coordinates. To be seen, trace `hoverinfo` must contain a "text" flag. """
	)
	ids: Optional[constr(pattern=STATEMENT_REGEX) | List]= Field(
		None,
		description=""" data array<br>Assigns id labels to each datum. These ids for object constancy of data points during animation. Should be an array of strings, not numbers or any other type. """
	)
	increasing: Optional[WaterfallIncreasing]= Field(
		None,
		description=""" object containing one or more of the keys listed below.<br> """
	)
	insidetextanchor: Optional[str]= Field(
		None,
		description=""" enumerated , one of ( "end" | "middle" | "start" )<br>Determines if texts are kept at center or start/end points in `textposition` "inside" mode. """
	)
	insidetextfont: Optional[FontInsidetextfontTextfontOutsidetextfont1]= Field(
		None,
		description=""" object containing one or more of the keys listed below.<br>Sets the font used for `text` lying inside the bar. """
	)
	legendgroup: Optional[str]= Field(
		None,
		description=""" string<br>Sets the legend group for this trace. Traces part of the same legend group hide/show at the same time when toggling legend items. """
	)
	legendgrouptitle: Optional[WaterfallLegendgrouptitle]= Field(
		None,
		description=""" object containing one or more of the keys listed below.<br> """
	)
	legendrank: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number<br>Sets the legend rank for this trace. Items and groups with smaller ranks are presented on top/left side while with `"reversed" `legend.traceorder` they are on bottom/right side. The default legendrank is 1000, so that you can use ranks less than 1000 to place certain items before all unranked items, and ranks greater than 1000 to go after all unranked items. """
	)
	legendwidth: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number greater than or equal to 0<br>Sets the width (in px or fraction) of the legend for this trace. """
	)
	measure: Optional[constr(pattern=STATEMENT_REGEX) | List]= Field(
		None,
		description=""" data array<br>An array containing types of values. By default the values are considered as 'relative'. However; it is possible to use 'total' to compute the sums. Also 'absolute' could be applied to reset the computed total or to declare an initial value where needed. """
	)
	meta: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number or categorical coordinate string<br>Assigns extra meta information associated with this trace that can be used in various text attributes. Attributes such as trace `name`, graph, axis and colorbar `title.text`, annotation `text` `rangeselector`, `updatemenues` and `sliders` `label` text all support `meta`. To access the trace `meta` values in an attribute in the same trace, simply use `%{meta[i]}` where `i` is the index or key of the `meta` item in question. To access trace `meta` in layout attributes, use `%{data[n[.meta[i]}` where `i` is the index or key of the `meta` and `n` is the trace index. """
	)
	offset: Optional[constr(pattern=INDEXED_STATEMENT_REGEX) | constr(pattern=STATEMENT_REGEX) | float | List[float]]= Field(
		None,
		description=""" number or array of numbers<br>Shifts the position where the bar is drawn (in position axis units). In "group" barmode, traces that set "offset" will be excluded and drawn in "overlay" mode instead. """
	)
	offsetgroup: Optional[str]= Field(
		None,
		description=""" string<br>Set several traces linked to the same position axis or matching axes to the same offsetgroup where bars of the same position coordinate will line up. """
	)
	opacity: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number between or equal to 0 and 1<br>Sets the opacity of the trace. """
	)
	orientation: Optional[str]= Field(
		None,
		description=""" enumerated , one of ( "v" | "h" )<br>Sets the orientation of the bars. With "v" ("h"), the value of the each bar spans along the vertical (horizontal). """
	)
	outsidetextfont: Optional[FontInsidetextfontTextfontOutsidetextfont1]= Field(
		None,
		description=""" object containing one or more of the keys listed below.<br>Sets the font used for `text` lying outside the bar. """
	)
	selectedpoints: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number or categorical coordinate string<br>Array containing integer indices of selected points. Has an effect only for traces that support selections. Note that an empty array means an empty selection where the `unselected` are turned on for all points, whereas, any other non-array values means no selection all where the `selected` and `unselected` styles have no effect. """
	)
	showlegend: Optional[bool | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" boolean<br>Determines whether or not an item corresponding to this trace is shown in the legend. """
	)
	text: Optional[str | List[str]]= Field(
		None,
		description=""" string or array of strings<br>Sets text elements associated with each (x,y) pair. If a single string, the same string appears over all the data points. If an array of string, the items are mapped in order to the this trace's (x,y) coordinates. If trace `hoverinfo` contains a "text" flag and "hovertext" is not set, these elements will be seen in the hover labels. """
	)
	textangle: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" angle<br>Sets the angle of the tick labels with respect to the bar. For example, a `tickangle` of -90 draws the tick labels vertically. With "auto" the texts may automatically be rotated to fit with the maximum size in bars. """
	)
	textfont: Optional[FontInsidetextfontTextfontOutsidetextfont1]= Field(
		None,
		description=""" object containing one or more of the keys listed below.<br>Sets the font used for `text`. """
	)
	textinfo: Optional[str]= Field(
		None,
		description=""" flaglist string. any combination of "label", "text", "initial", "delta", "final" joined with a "+" or "none".<br>Determines which trace information appear on the graph. In the case of having multiple waterfalls, totals are computed separately (per trace). """
	)
	textposition: Optional[str | List[str] ]= Field(
		None,
		description=""" enumerated or array of enumerateds , one of ( "inside" | "outside" | "auto" | "none" )<br>Specifies the location of the `text`. "inside" positions `text` inside, next to the bar end (rotated and scaled if needed). "outside" positions `text` outside, next to the bar end (scaled if needed), unless there is another bar stacked on this one, then the text gets pushed inside. "auto" tries to position `text` inside the bar, but if the bar is too small and no bar is stacked on this one the text is moved outside. If "none", no text appears. """
	)
	texttemplate: Optional[str | List[str]]= Field(
		None,
		description=""" string or array of strings<br>Template string used for rendering the information text that appear on points. Note that this will override `textinfo`. Variables are inserted using %{variable}, for example "y: %{y}". Numbers are formatted using d3-format's syntax %{variable:d3-format}, for example "Price: %{y:$.2f}". https://github.com/d3/d3-format/tree/v1.4.5#d3-format for details on the formatting syntax. Dates are formatted using d3-time-format's syntax %{variable|d3-time-format}, for example "Day: %{2019-01-01|%A}". https://github.com/d3/d3-time-format/tree/v2.2.3#locale_format for details on the date formatting syntax. Every attributes that can be specified per-point (the ones that are `arrayOk: true`) are available. variables `initial`, `delta`, `final` and `label`. """
	)
	totals: Optional[WaterfallTotals]= Field(
		None,
		description=""" object containing one or more of the keys listed below.<br> """
	)
	type: Literal["waterfall"]= Field(
		...,
		description=""" "waterfall"<br> """
	)
	uirevision: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number or categorical coordinate string<br>Controls persistence of some user-driven changes to the trace: `constraintrange` in `parcoords` traces, as well as some `editable: true` modifications such as `name` and `colorbar.title`. Defaults to `layout.uirevision`. Note that other user-driven trace attribute changes are controlled by `layout` attributes: `trace.visible` is controlled by `layout.legend.uirevision`, `selectedpoints` is controlled by `layout.selectionrevision`, and `colorbar.(x|y)` (accessible with `config: {editable: true}`) is controlled by `layout.editrevision`. Trace changes are tracked by `uid`, which only falls back on trace index if no `uid` is provided. So if your app can add/remove traces before the end of the `data` array, such that the same trace has a different index, you can still preserve user-driven changes if you give each trace a `uid` that stays with it as it moves. """
	)
	visible: Optional[str]= Field(
		None,
		description=""" enumerated , one of ( true | false | "legendonly" )<br>Determines whether or not this trace is visible. If "legendonly", the trace is not drawn, but can appear as a legend item (provided that the legend itself is visible). """
	)
	width: Optional[constr(pattern=INDEXED_STATEMENT_REGEX) | constr(pattern=STATEMENT_REGEX) | float | List[float]]= Field(
		None,
		description=""" number or array of numbers greater than or equal to 0<br>Sets the bar width (in position axis units). """
	)
	x0: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number or categorical coordinate string<br>Alternate to `x`. Builds a linear space of x coordinates. Use with `dx` where `x0` is the starting coordinate and `dx` the step. """
	)
	x: Optional[constr(pattern=STATEMENT_REGEX) | List]= Field(
		None,
		description=""" data array<br>Sets the x coordinates. """
	)
	xaxis: Optional[str]= Field(
		None,
		description=""" subplotid<br>Sets a reference between this trace's x coordinates and a 2D cartesian x axis. If "x" (the default value), the x coordinates refer to `layout.xaxis`. If "x2", the x coordinates refer to `layout.xaxis2`, and so on. """
	)
	xhoverformat: Optional[str]= Field(
		None,
		description=""" string<br>Sets the hover text formatting rulefor `x` using d3 formatting mini-languages which are very similar to those in Python. For numbers, see: https://github.com/d3/d3-format/tree/v1.4.5#d3-format. And for dates see: https://github.com/d3/d3-time-format/tree/v2.2.3#locale_format. We add two items to d3's date formatter: "%h" for half of the year as a decimal number as well as "%{n}f" for fractional seconds with n digits. For example, "2016-10-13 09:15:23.456" with tickformat "%H~%M~%S.%2f" would display "09~15~23.46"By default the values are formatted using `xaxis.hoverformat`. """
	)
	xperiod0: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number or categorical coordinate string<br>Only relevant when the axis `type` is "date". Sets the base for period positioning in milliseconds or date string on the x0 axis. When `x0period` is round number of weeks, the `x0period0` by default would be on a Sunday i.e. 2000-01-02, otherwise it would be at 2000-01-01. """
	)
	xperiod: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number or categorical coordinate string<br>Only relevant when the axis `type` is "date". Sets the period positioning in milliseconds or "M<n>" on the x axis. Special values in the form of "M<n>" could be used to declare the number of months. In this case `n` must be a positive integer. """
	)
	xperiodalignment: Optional[str]= Field(
		None,
		description=""" enumerated , one of ( "start" | "middle" | "end" )<br>Only relevant when the axis `type` is "date". Sets the alignment of data points on the x axis. """
	)
	y0: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number or categorical coordinate string<br>Alternate to `y`. Builds a linear space of y coordinates. Use with `dy` where `y0` is the starting coordinate and `dy` the step. """
	)
	y: Optional[constr(pattern=STATEMENT_REGEX) | List]= Field(
		None,
		description=""" data array<br>Sets the y coordinates. """
	)
	yaxis: Optional[str]= Field(
		None,
		description=""" subplotid<br>Sets a reference between this trace's y coordinates and a 2D cartesian y axis. If "y" (the default value), the y coordinates refer to `layout.yaxis`. If "y2", the y coordinates refer to `layout.yaxis2`, and so on. """
	)
	yhoverformat: Optional[str]= Field(
		None,
		description=""" string<br>Sets the hover text formatting rulefor `y` using d3 formatting mini-languages which are very similar to those in Python. For numbers, see: https://github.com/d3/d3-format/tree/v1.4.5#d3-format. And for dates see: https://github.com/d3/d3-time-format/tree/v2.2.3#locale_format. We add two items to d3's date formatter: "%h" for half of the year as a decimal number as well as "%{n}f" for fractional seconds with n digits. For example, "2016-10-13 09:15:23.456" with tickformat "%H~%M~%S.%2f" would display "09~15~23.46"By default the values are formatted using `yaxis.hoverformat`. """
	)
	yperiod0: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number or categorical coordinate string<br>Only relevant when the axis `type` is "date". Sets the base for period positioning in milliseconds or date string on the y0 axis. When `y0period` is round number of weeks, the `y0period0` by default would be on a Sunday i.e. 2000-01-02, otherwise it would be at 2000-01-01. """
	)
	yperiod: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number or categorical coordinate string<br>Only relevant when the axis `type` is "date". Sets the period positioning in milliseconds or "M<n>" on the y axis. Special values in the form of "M<n>" could be used to declare the number of months. In this case `n` must be a positive integer. """
	)
	yperiodalignment: Optional[str]= Field(
		None,
		description=""" enumerated , one of ( "start" | "middle" | "end" )<br>Only relevant when the axis `type` is "date". Sets the alignment of data points on the y axis. """
	)



from pydantic import Field, constr
from visivo.models.base.base_model import INDEXED_STATEMENT_REGEX, STATEMENT_REGEX
from visivo.models.trace_props.trace_props import  TraceProps, TracePropsAttribute
from typing import List, Literal, Optional 


class FontInsidetextfontTextfontOutsidetextfont1(TracePropsAttribute):
	color: Optional[str | List[str]]= Field(
		None,
		description=""" color or array of colors<br> """
	)
	family: Optional[str | List[str]]= Field(
		None,
		description=""" string or array of strings<br>HTML font family - the typeface that will be applied by the web browser. The web browser will only be able to apply a font if it is available on the system which it operates. Provide multiple font families, separated by commas, to indicate the preference in which to apply fonts if they aren't available on the system. The Chart Studio Cloud (at https://chart-studio.plotly.com or on-premise) generates images on a server, where only a select number of fonts are installed and supported. These include "Arial", "Balto", "Courier New", "Droid Sans",, "Droid Serif", "Droid Sans Mono", "Gravitas One", "Old Standard TT", "Open Sans", "Overpass", "PT Sans Narrow", "Raleway", "Times New Roman". """
	)
	size: Optional[constr(pattern=INDEXED_STATEMENT_REGEX) | constr(pattern=STATEMENT_REGEX) | float | List[float]]= Field(
		None,
		description=""" number or array of numbers greater than or equal to 1<br> """
	)
class BoxHoverlabel(TracePropsAttribute):
	align: Optional[str | List[str] ]= Field(
		None,
		description=""" enumerated or array of enumerateds , one of ( "left" | "right" | "auto" )<br>Sets the horizontal alignment of the text content within hover label box. Has an effect only if the hover label text spans more two or more lines """
	)
	bgcolor: Optional[str | List[str]]= Field(
		None,
		description=""" color or array of colors<br>Sets the background color of the hover labels for this trace """
	)
	bordercolor: Optional[str | List[str]]= Field(
		None,
		description=""" color or array of colors<br>Sets the border color of the hover labels for this trace. """
	)
	font: Optional[FontInsidetextfontTextfontOutsidetextfont1]= Field(
		None,
		description=""" object containing one or more of the keys listed below.<br>Sets the font used in hover labels. """
	)
	namelength: Optional[int | constr(pattern=INDEXED_STATEMENT_REGEX) | constr(pattern=STATEMENT_REGEX) | List[int]]= Field(
		None,
		description=""" integer or array of integers greater than or equal to -1<br>Sets the default length (in number of characters) of the trace name in the hover labels for all traces. -1 shows the whole name regardless of length. 0-3 shows the first 0-3 characters, and an integer >3 will show the whole name if it is less than that many characters, but if it is longer, will truncate to `namelength - 3` characters and add an ellipsis. """
	)
class TextfontLabelfontTickfontInsidetextfontRangefontOutsidetextfontFontGrouptitlefont1(TracePropsAttribute):
	color: Optional[str]= Field(
		None,
		description=""" color<br> """
	)
	family: Optional[str]= Field(
		None,
		description=""" string<br>HTML font family - the typeface that will be applied by the web browser. The web browser will only be able to apply a font if it is available on the system which it operates. Provide multiple font families, separated by commas, to indicate the preference in which to apply fonts if they aren't available on the system. The Chart Studio Cloud (at https://chart-studio.plotly.com or on-premise) generates images on a server, where only a select number of fonts are installed and supported. These include "Arial", "Balto", "Courier New", "Droid Sans",, "Droid Serif", "Droid Sans Mono", "Gravitas One", "Old Standard TT", "Open Sans", "Overpass", "PT Sans Narrow", "Raleway", "Times New Roman". """
	)
	size: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number greater than or equal to 1<br> """
	)
class BoxLegendgrouptitle(TracePropsAttribute):
	font: Optional[TextfontLabelfontTickfontInsidetextfontRangefontOutsidetextfontFontGrouptitlefont1]= Field(
		None,
		description=""" object containing one or more of the keys listed below.<br>Sets this legend group's title font. """
	)
	text: Optional[str]= Field(
		None,
		description=""" string<br>Sets the title of the legend group. """
	)
class Line5(TracePropsAttribute):
	color: Optional[str]= Field(
		None,
		description=""" color<br>Sets the color of line bounding the box(es). """
	)
	width: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number greater than or equal to 0<br>Sets the width (in px) of line bounding the box(es). """
	)
class Line8(TracePropsAttribute):
	color: Optional[str]= Field(
		None,
		description=""" color<br>Sets the marker.line color. It accepts either a specific color or an array of numbers that are mapped to the colorscale relative to the max and min values of the array or relative to `marker.line.cmin` and `marker.line.cmax` if set. """
	)
	outliercolor: Optional[str]= Field(
		None,
		description=""" color<br>Sets the border line color of the outlier sample points. Defaults to marker.color """
	)
	outlierwidth: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number greater than or equal to 0<br>Sets the border line width (in px) of the outlier sample points. """
	)
	width: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number greater than or equal to 0<br>Sets the width (in px) of the lines bounding the marker points. """
	)
class BoxMarker(TracePropsAttribute):
	angle: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" angle<br>Sets the marker angle in respect to `angleref`. """
	)
	color: Optional[str]= Field(
		None,
		description=""" color<br>Sets the marker color. It accepts either a specific color or an array of numbers that are mapped to the colorscale relative to the max and min values of the array or relative to `marker.cmin` and `marker.cmax` if set. """
	)
	line: Optional[Line8]= Field(
		None,
		description=""" object containing one or more of the keys listed below.<br> """
	)
	opacity: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number between or equal to 0 and 1<br>Sets the marker opacity. """
	)
	outliercolor: Optional[str]= Field(
		None,
		description=""" color<br>Sets the color of the outlier sample points. """
	)
	size: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number greater than or equal to 0<br>Sets the marker size (in px). """
	)
	symbol: Optional[str]= Field(
		None,
		description=""" enumerated , one of ( "0" | "0" | "circle" | "100" | "100" | "circle-open" | "200" | "200" | "circle-dot" | "300" | "300" | "circle-open-dot" | "1" | "1" | "square" | "101" | "101" | "square-open" | "201" | "201" | "square-dot" | "301" | "301" | "square-open-dot" | "2" | "2" | "diamond" | "102" | "102" | "diamond-open" | "202" | "202" | "diamond-dot" | "302" | "302" | "diamond-open-dot" | "3" | "3" | "cross" | "103" | "103" | "cross-open" | "203" | "203" | "cross-dot" | "303" | "303" | "cross-open-dot" | "4" | "4" | "x" | "104" | "104" | "x-open" | "204" | "204" | "x-dot" | "304" | "304" | "x-open-dot" | "5" | "5" | "triangle-up" | "105" | "105" | "triangle-up-open" | "205" | "205" | "triangle-up-dot" | "305" | "305" | "triangle-up-open-dot" | "6" | "6" | "triangle-down" | "106" | "106" | "triangle-down-open" | "206" | "206" | "triangle-down-dot" | "306" | "306" | "triangle-down-open-dot" | "7" | "7" | "triangle-left" | "107" | "107" | "triangle-left-open" | "207" | "207" | "triangle-left-dot" | "307" | "307" | "triangle-left-open-dot" | "8" | "8" | "triangle-right" | "108" | "108" | "triangle-right-open" | "208" | "208" | "triangle-right-dot" | "308" | "308" | "triangle-right-open-dot" | "9" | "9" | "triangle-ne" | "109" | "109" | "triangle-ne-open" | "209" | "209" | "triangle-ne-dot" | "309" | "309" | "triangle-ne-open-dot" | "10" | "10" | "triangle-se" | "110" | "110" | "triangle-se-open" | "210" | "210" | "triangle-se-dot" | "310" | "310" | "triangle-se-open-dot" | "11" | "11" | "triangle-sw" | "111" | "111" | "triangle-sw-open" | "211" | "211" | "triangle-sw-dot" | "311" | "311" | "triangle-sw-open-dot" | "12" | "12" | "triangle-nw" | "112" | "112" | "triangle-nw-open" | "212" | "212" | "triangle-nw-dot" | "312" | "312" | "triangle-nw-open-dot" | "13" | "13" | "pentagon" | "113" | "113" | "pentagon-open" | "213" | "213" | "pentagon-dot" | "313" | "313" | "pentagon-open-dot" | "14" | "14" | "hexagon" | "114" | "114" | "hexagon-open" | "214" | "214" | "hexagon-dot" | "314" | "314" | "hexagon-open-dot" | "15" | "15" | "hexagon2" | "115" | "115" | "hexagon2-open" | "215" | "215" | "hexagon2-dot" | "315" | "315" | "hexagon2-open-dot" | "16" | "16" | "octagon" | "116" | "116" | "octagon-open" | "216" | "216" | "octagon-dot" | "316" | "316" | "octagon-open-dot" | "17" | "17" | "star" | "117" | "117" | "star-open" | "217" | "217" | "star-dot" | "317" | "317" | "star-open-dot" | "18" | "18" | "hexagram" | "118" | "118" | "hexagram-open" | "218" | "218" | "hexagram-dot" | "318" | "318" | "hexagram-open-dot" | "19" | "19" | "star-triangle-up" | "119" | "119" | "star-triangle-up-open" | "219" | "219" | "star-triangle-up-dot" | "319" | "319" | "star-triangle-up-open-dot" | "20" | "20" | "star-triangle-down" | "120" | "120" | "star-triangle-down-open" | "220" | "220" | "star-triangle-down-dot" | "320" | "320" | "star-triangle-down-open-dot" | "21" | "21" | "star-square" | "121" | "121" | "star-square-open" | "221" | "221" | "star-square-dot" | "321" | "321" | "star-square-open-dot" | "22" | "22" | "star-diamond" | "122" | "122" | "star-diamond-open" | "222" | "222" | "star-diamond-dot" | "322" | "322" | "star-diamond-open-dot" | "23" | "23" | "diamond-tall" | "123" | "123" | "diamond-tall-open" | "223" | "223" | "diamond-tall-dot" | "323" | "323" | "diamond-tall-open-dot" | "24" | "24" | "diamond-wide" | "124" | "124" | "diamond-wide-open" | "224" | "224" | "diamond-wide-dot" | "324" | "324" | "diamond-wide-open-dot" | "25" | "25" | "hourglass" | "125" | "125" | "hourglass-open" | "26" | "26" | "bowtie" | "126" | "126" | "bowtie-open" | "27" | "27" | "circle-cross" | "127" | "127" | "circle-cross-open" | "28" | "28" | "circle-x" | "128" | "128" | "circle-x-open" | "29" | "29" | "square-cross" | "129" | "129" | "square-cross-open" | "30" | "30" | "square-x" | "130" | "130" | "square-x-open" | "31" | "31" | "diamond-cross" | "131" | "131" | "diamond-cross-open" | "32" | "32" | "diamond-x" | "132" | "132" | "diamond-x-open" | "33" | "33" | "cross-thin" | "133" | "133" | "cross-thin-open" | "34" | "34" | "x-thin" | "134" | "134" | "x-thin-open" | "35" | "35" | "asterisk" | "135" | "135" | "asterisk-open" | "36" | "36" | "hash" | "136" | "136" | "hash-open" | "236" | "236" | "hash-dot" | "336" | "336" | "hash-open-dot" | "37" | "37" | "y-up" | "137" | "137" | "y-up-open" | "38" | "38" | "y-down" | "138" | "138" | "y-down-open" | "39" | "39" | "y-left" | "139" | "139" | "y-left-open" | "40" | "40" | "y-right" | "140" | "140" | "y-right-open" | "41" | "41" | "line-ew" | "141" | "141" | "line-ew-open" | "42" | "42" | "line-ns" | "142" | "142" | "line-ns-open" | "43" | "43" | "line-ne" | "143" | "143" | "line-ne-open" | "44" | "44" | "line-nw" | "144" | "144" | "line-nw-open" | "45" | "45" | "arrow-up" | "145" | "145" | "arrow-up-open" | "46" | "46" | "arrow-down" | "146" | "146" | "arrow-down-open" | "47" | "47" | "arrow-left" | "147" | "147" | "arrow-left-open" | "48" | "48" | "arrow-right" | "148" | "148" | "arrow-right-open" | "49" | "49" | "arrow-bar-up" | "149" | "149" | "arrow-bar-up-open" | "50" | "50" | "arrow-bar-down" | "150" | "150" | "arrow-bar-down-open" | "51" | "51" | "arrow-bar-left" | "151" | "151" | "arrow-bar-left-open" | "52" | "52" | "arrow-bar-right" | "152" | "152" | "arrow-bar-right-open" | "53" | "53" | "arrow" | "153" | "153" | "arrow-open" | "54" | "54" | "arrow-wide" | "154" | "154" | "arrow-wide-open" )<br>Sets the marker symbol type. Adding 100 is equivalent to appending "-open" to a symbol name. Adding 200 is equivalent to appending "-dot" to a symbol name. Adding 300 is equivalent to appending "-open-dot" or "dot-open" to a symbol name. """
	)
class Marker1(TracePropsAttribute):
	color: Optional[str]= Field(
		None,
		description=""" color<br>Sets the marker color of selected points. """
	)
	opacity: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number between or equal to 0 and 1<br>Sets the marker opacity of selected points. """
	)
	size: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number greater than or equal to 0<br>Sets the marker size of selected points. """
	)
class BoxSelected(TracePropsAttribute):
	marker: Optional[Marker1]= Field(
		None,
		description=""" object containing one or more of the keys listed below.<br> """
	)
class Marker2(TracePropsAttribute):
	color: Optional[str]= Field(
		None,
		description=""" color<br>Sets the marker color of unselected points, applied only when a selection exists. """
	)
	opacity: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number between or equal to 0 and 1<br>Sets the marker opacity of unselected points, applied only when a selection exists. """
	)
	size: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number greater than or equal to 0<br>Sets the marker size of unselected points, applied only when a selection exists. """
	)
class BoxUnselected(TracePropsAttribute):
	marker: Optional[Marker2]= Field(
		None,
		description=""" object containing one or more of the keys listed below.<br> """
	)
class Box(TraceProps):
	alignmentgroup: Optional[str]= Field(
		None,
		description=""" string<br>Set several traces linked to the same position axis or matching axes to the same alignmentgroup. This controls whether bars compute their positional range dependently or independently. """
	)
	boxmean: Optional[str]= Field(
		None,
		description=""" enumerated , one of ( true | "sd" | false )<br>If "true", the mean of the box(es)' underlying distribution is drawn as a dashed line inside the box(es). If "sd" the standard deviation is also drawn. Defaults to "true" when `mean` is set. Defaults to "sd" when `sd` is set Otherwise defaults to "false". """
	)
	boxpoints: Optional[str]= Field(
		None,
		description=""" enumerated , one of ( "all" | "outliers" | "suspectedoutliers" | false )<br>If "outliers", only the sample points lying outside the whiskers are shown If "suspectedoutliers", the outlier points are shown and points either less than 4"Q1-3"Q3 or greater than 4"Q3-3"Q1 are highlighted (see `outliercolor`) If "all", all sample points are shown If "false", only the box(es) are shown with no sample points Defaults to "suspectedoutliers" when `marker.outliercolor` or `marker.line.outliercolor` is set. Defaults to "all" under the q1/median/q3 signature. Otherwise defaults to "outliers". """
	)
	customdata: Optional[constr(pattern=STATEMENT_REGEX) | List]= Field(
		None,
		description=""" data array<br>Assigns extra data each datum. This may be useful when listening to hover, click and selection events. Note that, "scatter" traces also appends customdata items in the markers DOM elements """
	)
	dx: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number<br>Sets the x coordinate step for multi-box traces set using q1/median/q3. """
	)
	dy: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number<br>Sets the y coordinate step for multi-box traces set using q1/median/q3. """
	)
	fillcolor: Optional[str]= Field(
		None,
		description=""" color<br>Sets the fill color. Defaults to a half-transparent variant of the line color, marker color, or marker line color, whichever is available. """
	)
	hoverinfo: Optional[str]= Field(
		None,
		description=""" flaglist string. any combination of "x", "y", "z", "text", "name" joined with a "+" or "all" or "none" or "skip".<br>Determines which trace information appear on hover. If `none` or `skip` are set, no information is displayed upon hovering. But, if `none` is set, click and hover events are still fired. """
	)
	hoverlabel: Optional[BoxHoverlabel]= Field(
		None,
		description=""" object containing one or more of the keys listed below.<br> """
	)
	hoveron: Optional[str]= Field(
		None,
		description=""" flaglist string. any combination of "boxes", "points" joined with a "+"<br>Do the hover effects highlight individual boxes or sample points or both? """
	)
	hovertemplate: Optional[str | List[str]]= Field(
		None,
		description=""" string or array of strings<br>Template string used for rendering the information that appear on hover box. Note that this will override `hoverinfo`. Variables are inserted using %{variable}, for example "y: %{y}" as well as %{xother}, {%_xother}, {%_xother_}, {%xother_}. When showing info for several points, "xother" will be added to those with different x positions from the first point. An underscore before or after "(x|y)other" will add a space on that side, only when this field is shown. Numbers are formatted using d3-format's syntax %{variable:d3-format}, for example "Price: %{y:$.2f}". https://github.com/d3/d3-format/tree/v1.4.5#d3-format for details on the formatting syntax. Dates are formatted using d3-time-format's syntax %{variable|d3-time-format}, for example "Day: %{2019-01-01|%A}". https://github.com/d3/d3-time-format/tree/v2.2.3#locale_format for details on the date formatting syntax. The variables available in `hovertemplate` are the ones emitted as event data described at this link https://plotly.com/javascript/plotlyjs-events/#event-data. Additionally, every attributes that can be specified per-point (the ones that are `arrayOk: true`) are available. Anything contained in tag `<extra>` is displayed in the secondary box, for example "<extra>{fullData.name}</extra>". To hide the secondary box completely, use an empty tag `<extra></extra>`. """
	)
	hovertext: Optional[str | List[str]]= Field(
		None,
		description=""" string or array of strings<br>Same as `text`. """
	)
	ids: Optional[constr(pattern=STATEMENT_REGEX) | List]= Field(
		None,
		description=""" data array<br>Assigns id labels to each datum. These ids for object constancy of data points during animation. Should be an array of strings, not numbers or any other type. """
	)
	jitter: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number between or equal to 0 and 1<br>Sets the amount of jitter in the sample points drawn. If "0", the sample points align along the distribution axis. If "1", the sample points are drawn in a random jitter of width equal to the width of the box(es). """
	)
	legendgroup: Optional[str]= Field(
		None,
		description=""" string<br>Sets the legend group for this trace. Traces part of the same legend group hide/show at the same time when toggling legend items. """
	)
	legendgrouptitle: Optional[BoxLegendgrouptitle]= Field(
		None,
		description=""" object containing one or more of the keys listed below.<br> """
	)
	legendrank: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number<br>Sets the legend rank for this trace. Items and groups with smaller ranks are presented on top/left side while with `"reversed" `legend.traceorder` they are on bottom/right side. The default legendrank is 1000, so that you can use ranks less than 1000 to place certain items before all unranked items, and ranks greater than 1000 to go after all unranked items. """
	)
	legendwidth: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number greater than or equal to 0<br>Sets the width (in px or fraction) of the legend for this trace. """
	)
	line: Optional[Line5]= Field(
		None,
		description=""" object containing one or more of the keys listed below.<br> """
	)
	lowerfence: Optional[constr(pattern=STATEMENT_REGEX) | List]= Field(
		None,
		description=""" data array<br>Sets the lower fence values. There should be as many items as the number of boxes desired. This attribute has effect only under the q1/median/q3 signature. If `lowerfence` is not provided but a sample (in `y` or `x`) is set, we compute the lower as the last sample point below 1.5 times the IQR. """
	)
	marker: Optional[BoxMarker]= Field(
		None,
		description=""" object containing one or more of the keys listed below.<br> """
	)
	mean: Optional[constr(pattern=STATEMENT_REGEX) | List]= Field(
		None,
		description=""" data array<br>Sets the mean values. There should be as many items as the number of boxes desired. This attribute has effect only under the q1/median/q3 signature. If `mean` is not provided but a sample (in `y` or `x`) is set, we compute the mean for each box using the sample values. """
	)
	median: Optional[constr(pattern=STATEMENT_REGEX) | List]= Field(
		None,
		description=""" data array<br>Sets the median values. There should be as many items as the number of boxes desired. """
	)
	meta: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number or categorical coordinate string<br>Assigns extra meta information associated with this trace that can be used in various text attributes. Attributes such as trace `name`, graph, axis and colorbar `title.text`, annotation `text` `rangeselector`, `updatemenues` and `sliders` `label` text all support `meta`. To access the trace `meta` values in an attribute in the same trace, simply use `%{meta[i]}` where `i` is the index or key of the `meta` item in question. To access trace `meta` in layout attributes, use `%{data[n[.meta[i]}` where `i` is the index or key of the `meta` and `n` is the trace index. """
	)
	notched: Optional[bool | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" boolean<br>Determines whether or not notches are drawn. Notches displays a confidence interval around the median. We compute the confidence interval as median +/- 1.57 " IQR / sqrt(N), where IQR is the interquartile range and N is the sample size. If two boxes' notches do not overlap there is 95% confidence their medians differ. See https://sites.google.com/site/davidsstatistics/home/notched-box-plots for more info. Defaults to "false" unless `notchwidth` or `notchspan` is set. """
	)
	notchspan: Optional[constr(pattern=STATEMENT_REGEX) | List]= Field(
		None,
		description=""" data array<br>Sets the notch span from the boxes' `median` values. There should be as many items as the number of boxes desired. This attribute has effect only under the q1/median/q3 signature. If `notchspan` is not provided but a sample (in `y` or `x`) is set, we compute it as 1.57 " IQR / sqrt(N), where N is the sample size. """
	)
	notchwidth: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number between or equal to 0 and 0.5<br>Sets the width of the notches relative to the box' width. For example, with 0, the notches are as wide as the box(es). """
	)
	offsetgroup: Optional[str]= Field(
		None,
		description=""" string<br>Set several traces linked to the same position axis or matching axes to the same offsetgroup where bars of the same position coordinate will line up. """
	)
	opacity: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number between or equal to 0 and 1<br>Sets the opacity of the trace. """
	)
	orientation: Optional[str]= Field(
		None,
		description=""" enumerated , one of ( "v" | "h" )<br>Sets the orientation of the box(es). If "v" ("h"), the distribution is visualized along the vertical (horizontal). """
	)
	pointpos: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number between or equal to -2 and 2<br>Sets the position of the sample points in relation to the box(es). If "0", the sample points are places over the center of the box(es). Positive (negative) values correspond to positions to the right (left) for vertical boxes and above (below) for horizontal boxes """
	)
	q1: Optional[constr(pattern=STATEMENT_REGEX) | List]= Field(
		None,
		description=""" data array<br>Sets the Quartile 1 values. There should be as many items as the number of boxes desired. """
	)
	q3: Optional[constr(pattern=STATEMENT_REGEX) | List]= Field(
		None,
		description=""" data array<br>Sets the Quartile 3 values. There should be as many items as the number of boxes desired. """
	)
	quartilemethod: Optional[str]= Field(
		None,
		description=""" enumerated , one of ( "linear" | "exclusive" | "inclusive" )<br>Sets the method used to compute the sample's Q1 and Q3 quartiles. The "linear" method uses the 25th percentile for Q1 and 75th percentile for Q3 as computed using method #10 (listed on http://jse.amstat.org/v14n3/langford.html). The "exclusive" method uses the median to divide the ordered dataset into two halves if the sample is odd, it does not include the median in either half - Q1 is then the median of the lower half and Q3 the median of the upper half. The "inclusive" method also uses the median to divide the ordered dataset into two halves but if the sample is odd, it includes the median in both halves - Q1 is then the median of the lower half and Q3 the median of the upper half. """
	)
	sd: Optional[constr(pattern=STATEMENT_REGEX) | List]= Field(
		None,
		description=""" data array<br>Sets the standard deviation values. There should be as many items as the number of boxes desired. This attribute has effect only under the q1/median/q3 signature. If `sd` is not provided but a sample (in `y` or `x`) is set, we compute the standard deviation for each box using the sample values. """
	)
	selected: Optional[BoxSelected]= Field(
		None,
		description=""" object containing one or more of the keys listed below.<br> """
	)
	selectedpoints: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number or categorical coordinate string<br>Array containing integer indices of selected points. Has an effect only for traces that support selections. Note that an empty array means an empty selection where the `unselected` are turned on for all points, whereas, any other non-array values means no selection all where the `selected` and `unselected` styles have no effect. """
	)
	showlegend: Optional[bool | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" boolean<br>Determines whether or not an item corresponding to this trace is shown in the legend. """
	)
	text: Optional[str | List[str]]= Field(
		None,
		description=""" string or array of strings<br>Sets the text elements associated with each sample value. If a single string, the same string appears over all the data points. If an array of string, the items are mapped in order to the this trace's (x,y) coordinates. To be seen, trace `hoverinfo` must contain a "text" flag. """
	)
	type: Literal["box"]= Field(
		...,
		description=""" "box"<br> """
	)
	uirevision: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number or categorical coordinate string<br>Controls persistence of some user-driven changes to the trace: `constraintrange` in `parcoords` traces, as well as some `editable: true` modifications such as `name` and `colorbar.title`. Defaults to `layout.uirevision`. Note that other user-driven trace attribute changes are controlled by `layout` attributes: `trace.visible` is controlled by `layout.legend.uirevision`, `selectedpoints` is controlled by `layout.selectionrevision`, and `colorbar.(x|y)` (accessible with `config: {editable: true}`) is controlled by `layout.editrevision`. Trace changes are tracked by `uid`, which only falls back on trace index if no `uid` is provided. So if your app can add/remove traces before the end of the `data` array, such that the same trace has a different index, you can still preserve user-driven changes if you give each trace a `uid` that stays with it as it moves. """
	)
	unselected: Optional[BoxUnselected]= Field(
		None,
		description=""" object containing one or more of the keys listed below.<br> """
	)
	upperfence: Optional[constr(pattern=STATEMENT_REGEX) | List]= Field(
		None,
		description=""" data array<br>Sets the upper fence values. There should be as many items as the number of boxes desired. This attribute has effect only under the q1/median/q3 signature. If `upperfence` is not provided but a sample (in `y` or `x`) is set, we compute the lower as the last sample point above 1.5 times the IQR. """
	)
	visible: Optional[str]= Field(
		None,
		description=""" enumerated , one of ( true | false | "legendonly" )<br>Determines whether or not this trace is visible. If "legendonly", the trace is not drawn, but can appear as a legend item (provided that the legend itself is visible). """
	)
	whiskerwidth: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number between or equal to 0 and 1<br>Sets the width of the whiskers relative to the box' width. For example, with 1, the whiskers are as wide as the box(es). """
	)
	width: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number greater than or equal to 0<br>Sets the width of the box in data coordinate If "0" (default value) the width is automatically selected based on the positions of other box traces in the same subplot. """
	)
	x0: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number or categorical coordinate string<br>Sets the x coordinate for single-box traces or the starting coordinate for multi-box traces set using q1/median/q3. See overview for more info. """
	)
	x: Optional[constr(pattern=STATEMENT_REGEX) | List]= Field(
		None,
		description=""" data array<br>Sets the x sample data or coordinates. See overview for more info. """
	)
	xaxis: Optional[str]= Field(
		None,
		description=""" subplotid<br>Sets a reference between this trace's x coordinates and a 2D cartesian x axis. If "x" (the default value), the x coordinates refer to `layout.xaxis`. If "x2", the x coordinates refer to `layout.xaxis2`, and so on. """
	)
	xcalendar: Optional[str]= Field(
		None,
		description=""" enumerated , one of ( "chinese" | "coptic" | "discworld" | "ethiopian" | "gregorian" | "hebrew" | "islamic" | "jalali" | "julian" | "mayan" | "nanakshahi" | "nepali" | "persian" | "taiwan" | "thai" | "ummalqura" )<br>Sets the calendar system to use with `x` date data. """
	)
	xhoverformat: Optional[str]= Field(
		None,
		description=""" string<br>Sets the hover text formatting rulefor `x` using d3 formatting mini-languages which are very similar to those in Python. For numbers, see: https://github.com/d3/d3-format/tree/v1.4.5#d3-format. And for dates see: https://github.com/d3/d3-time-format/tree/v2.2.3#locale_format. We add two items to d3's date formatter: "%h" for half of the year as a decimal number as well as "%{n}f" for fractional seconds with n digits. For example, "2016-10-13 09:15:23.456" with tickformat "%H~%M~%S.%2f" would display "09~15~23.46"By default the values are formatted using `xaxis.hoverformat`. """
	)
	xperiod0: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number or categorical coordinate string<br>Only relevant when the axis `type` is "date". Sets the base for period positioning in milliseconds or date string on the x0 axis. When `x0period` is round number of weeks, the `x0period0` by default would be on a Sunday i.e. 2000-01-02, otherwise it would be at 2000-01-01. """
	)
	xperiod: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number or categorical coordinate string<br>Only relevant when the axis `type` is "date". Sets the period positioning in milliseconds or "M<n>" on the x axis. Special values in the form of "M<n>" could be used to declare the number of months. In this case `n` must be a positive integer. """
	)
	xperiodalignment: Optional[str]= Field(
		None,
		description=""" enumerated , one of ( "start" | "middle" | "end" )<br>Only relevant when the axis `type` is "date". Sets the alignment of data points on the x axis. """
	)
	y0: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number or categorical coordinate string<br>Sets the y coordinate for single-box traces or the starting coordinate for multi-box traces set using q1/median/q3. See overview for more info. """
	)
	y: Optional[constr(pattern=STATEMENT_REGEX) | List]= Field(
		None,
		description=""" data array<br>Sets the y sample data or coordinates. See overview for more info. """
	)
	yaxis: Optional[str]= Field(
		None,
		description=""" subplotid<br>Sets a reference between this trace's y coordinates and a 2D cartesian y axis. If "y" (the default value), the y coordinates refer to `layout.yaxis`. If "y2", the y coordinates refer to `layout.yaxis2`, and so on. """
	)
	ycalendar: Optional[str]= Field(
		None,
		description=""" enumerated , one of ( "chinese" | "coptic" | "discworld" | "ethiopian" | "gregorian" | "hebrew" | "islamic" | "jalali" | "julian" | "mayan" | "nanakshahi" | "nepali" | "persian" | "taiwan" | "thai" | "ummalqura" )<br>Sets the calendar system to use with `y` date data. """
	)
	yhoverformat: Optional[str]= Field(
		None,
		description=""" string<br>Sets the hover text formatting rulefor `y` using d3 formatting mini-languages which are very similar to those in Python. For numbers, see: https://github.com/d3/d3-format/tree/v1.4.5#d3-format. And for dates see: https://github.com/d3/d3-time-format/tree/v2.2.3#locale_format. We add two items to d3's date formatter: "%h" for half of the year as a decimal number as well as "%{n}f" for fractional seconds with n digits. For example, "2016-10-13 09:15:23.456" with tickformat "%H~%M~%S.%2f" would display "09~15~23.46"By default the values are formatted using `yaxis.hoverformat`. """
	)
	yperiod0: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number or categorical coordinate string<br>Only relevant when the axis `type` is "date". Sets the base for period positioning in milliseconds or date string on the y0 axis. When `y0period` is round number of weeks, the `y0period0` by default would be on a Sunday i.e. 2000-01-02, otherwise it would be at 2000-01-01. """
	)
	yperiod: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number or categorical coordinate string<br>Only relevant when the axis `type` is "date". Sets the period positioning in milliseconds or "M<n>" on the y axis. Special values in the form of "M<n>" could be used to declare the number of months. In this case `n` must be a positive integer. """
	)
	yperiodalignment: Optional[str]= Field(
		None,
		description=""" enumerated , one of ( "start" | "middle" | "end" )<br>Only relevant when the axis `type` is "date". Sets the alignment of data points on the y axis. """
	)


from typing import Annotated, Any, Union
from pydantic import Discriminator, Tag
from visivo.models.base.base_model import RefString
from visivo.models.targets.mysql_target import MysqlTarget
from visivo.models.targets.postgresql_target import PostgresqlTarget
from visivo.models.targets.snowflake_target import SnowflakeTarget
from visivo.models.targets.sqlite_target import SqliteTarget


def get_model_discriminator_value(value: Any) -> str:
    if isinstance(value, str):
        return "Ref"
    if isinstance(value, dict):
        if "type" in value:
            return value["type"]
    if hasattr(value, "type"):
        return value.type

    return None


TargetField = Annotated[
    Union[
        Annotated[SqliteTarget, Tag("sqlite")],
        Annotated[PostgresqlTarget, Tag("postgresql")],
        Annotated[MysqlTarget, Tag("mysql")],
        Annotated[SnowflakeTarget, Tag("snowflake")],
    ],
    Discriminator(get_model_discriminator_value),
]

TargetRefField = Annotated[
    Union[
        RefString,
        Annotated[SqliteTarget, Tag("sqlite")],
        Annotated[PostgresqlTarget, Tag("postgresql")],
        Annotated[MysqlTarget, Tag("mysql")],
        Annotated[SnowflakeTarget, Tag("snowflake")],
    ],
    Discriminator(get_model_discriminator_value),
]



from typing import Literal, Optional
from visivo.models.targets.sqlalchemy_target import SqlalchemyTarget
from pydantic import Field

MysqlType = Literal["mysql"]


class MysqlTarget(SqlalchemyTarget):
    """
    MysqlTargets hold the connection information to MySQL data sources.

    !!! example

        === "Simple"

            ``` yaml
                targets:
                  - name: mysql_target
                    type: mysql
                    database: database
                    username: {% raw %}{{ env_var('MYSQL_USER') }}{% endraw %}
                    password: {% raw %}{{ env_var('MYSQL_PASSWORD') }}{% endraw %}
                    connection_pool_size: 2
            ```

    !!! note

        Recommended environment variable use is covered in the [targets overview.](/topics/targets/)
    """

    type: MysqlType
    connection_pool_size: Optional[int] = Field(
        1, description="The pool size that is used for this connection."
    )

    def get_dialect(self):
        return "mysql+pymysql"



import click
from visivo.commands.options import (
    dist_dir,
    name_filter,
    output_dir,
    target,
    threads,
    working_dir,
)


@click.command()
@target
@working_dir
@output_dir
@dist_dir
@name_filter
@threads
def dist(output_dir, working_dir, dist_dir, target, name_filter, threads):
    """
    Creates a distributable version of this dashboard and stores it in a 'dist' folder.
    This folder can be statically deployed to any web server, hosting service, or bucket storage.
    """
    from visivo.logging.logger import Logger

    Logger.instance().debug("Creating Dist")

    from visivo.commands.dist_phase import dist_phase

    dist_phase(
        output_dir=output_dir,
        working_dir=working_dir,
        dist_dir=dist_dir,
        default_target=target,
        name_filter=name_filter,
        threads=threads,
    )
    Logger.instance().success(f"Created dist folder")



def _get_ref(field_data):
    refs = []

    # Check for ref inside 'items'
    ref = field_data.get("items", {}).get("$ref")
    if ref:
        refs.append(ref)
        return refs, "item"

    # Check for refs inside 'anyOf'
    any_of = field_data.get("anyOf", [])
    if any_of:
        for entry in any_of:
            ref = entry.get("$ref")
            if ref:
                refs.append(ref)
        return refs, "anyOf"

    # Check for refs inside 'oneOf'
    one_of = field_data.get("items", {}).get("oneOf", [])
    if one_of:
        for entry in one_of:
            ref = entry.get("$ref")
            if ref:
                refs.append(ref)
        return refs, "oneOf"

    # Check for refs inside 'discriminator' -> 'mapping'
    mapping_refs = list(field_data.get("discriminator", {}).get("mapping", {}).values())
    if mapping_refs:
        refs.extend(filter(None, mapping_refs))
        return refs, "mapping"

    return refs, None


def _process_model(schema, model_data, processed_models):

    properties = model_data.get("properties", {})
    nested_structure = {}

    for field, field_data in sorted(properties.items()):
        refs, ref_type = _get_ref(field_data)
        if refs and field not in ["props", "layout"] and ref_type != "oneOf":
            for ref in refs:
                nested_model_name = ref.split("/")[-1]
                nested_model_data = schema.get("$defs", {}).get(nested_model_name, {})
                nested_structure[nested_model_name] = _process_model(
                    schema, nested_model_data, processed_models
                )
        elif refs and field in ["props", "layout"]:
            for ref in refs:
                nested_model_name = ref.split("/")[-1]
                nested_model_data = schema.get("$defs", {}).get(nested_model_name, {})
                nested_structure[nested_model_name] = {}

        elif refs and ref_type == "oneOf":
            nested_structure[field.capitalize()] = {}
            for ref in refs:
                nested_model_name = ref.split("/")[-1]
                nested_model_data = schema.get("$defs", {}).get(nested_model_name, {})
                nested_structure[field.capitalize()][
                    nested_model_name
                ] = _process_model(schema, nested_model_data, processed_models)
        else:
            nested_structure[field] = field_data.get("type", "unknown")

    return nested_structure


def _generate_structure(schema):
    return _process_model(schema, schema, set())


def _to_mkdocs_yaml(schema, structure, base_path="reference/configuration"):
    output = []

    for model, contents in structure.items():
        if isinstance(contents, dict):
            sub_path = f"{base_path}/{model}"
            if schema.get("$defs", {}).get(model, {}):
                model_content = {
                    model: [f"{sub_path}/index.md"]
                    + _to_mkdocs_yaml(schema, contents, sub_path)
                }
            else:
                model_content = {model: _to_mkdocs_yaml(schema, contents, sub_path)}
            output.append(model_content)
        # Handle other cases if necessary (e.g., primitive types)

    return output


def _pop_nested_path(dictionary: dict, path: list):
    """Modifies dictionary removing key from a nested path, ignoring nested paths that may have already been removed"""
    result = dictionary
    for key in path[:-1]:
        if isinstance(result, dict) and key in result.keys():
            result = result[key]

    if path[-1] in result.keys():
        result.pop(path[-1])


def _pop_list_of_nested_paths(dictionary: dict, paths: list):
    for path in paths:
        _pop_nested_path(dictionary, path)


def _get_all_key_paths(dictionary, path=None) -> list:
    items = []
    path = path or []
    for key, value in dictionary.items():
        new_path = path + [key]
        if isinstance(value, dict):
            items.append({key: new_path})
            items += _get_all_key_paths(value, new_path)

    return items


def _consolidate_paths(paths: list) -> dict:
    consolidated = {}
    for dictionary in paths:
        model = list(dictionary.keys())[0]
        path = list(dictionary.values())[0]
        if model not in consolidated.keys():
            consolidated[model] = [path]
        elif model in consolidated.keys():
            consolidated[model].append(path)
    return consolidated


def _get_paths_to_remove(consolidated_paths: dict) -> list:
    to_remove = []
    for model, paths in consolidated_paths.items():

        top_path_length = 9999
        # find the shortest path to model
        for path in paths:
            if len(path) < top_path_length:
                top_path_length = len(path)
        # add longer paths to to_remove_list
        for path in paths:
            if len(path) != top_path_length:
                to_remove.append(path)
    return to_remove


def mkdocs_pydantic_nav(schema: dict) -> list:
    nested_structure = _generate_structure(schema)
    all_key_paths = _get_all_key_paths(nested_structure)
    consolidated_paths = _consolidate_paths(all_key_paths)
    paths_to_remove = _get_paths_to_remove(consolidated_paths)
    _pop_list_of_nested_paths(nested_structure, paths=paths_to_remove)
    yaml_output = _to_mkdocs_yaml(schema, nested_structure)
    return yaml_output


def _extract_strings_from_yaml(yaml_obj):
    result = []

    if isinstance(yaml_obj, str):
        return [yaml_obj]

    if isinstance(yaml_obj, list):
        for item in yaml_obj:
            result.extend(_extract_strings_from_yaml(item))

    if isinstance(yaml_obj, dict):
        for key, value in yaml_obj.items():
            result.extend(_extract_strings_from_yaml(value))

    return result


def get_model_to_page_mapping(nav_configuration: list) -> dict:
    file_paths = _extract_strings_from_yaml(nav_configuration)
    mapping = {}
    for path in file_paths:
        model = path.split("/")[-2]
        key = "#/$defs/" + model
        markdown_link = (
            f"[{model}]" + "(https://docs.visivo.io/" + path.replace("index.md", ")")
        )
        mapping[key] = markdown_link
    return mapping


def get_model_to_path_mapping(nav_configuration: list) -> dict:
    file_paths = _extract_strings_from_yaml(nav_configuration)
    mapping = {}
    for path in file_paths:
        model = path.split("/")[-2]
        file_path = "mkdocs/" + path
        mapping[model] = file_path
    return mapping


def find_path(object, key, path=None):
    """Gets the path to traverse through a nested dictionary list to get to a key. Finds the first key that matches."""
    if path is None:
        path = []

    if isinstance(object, dict):
        for k, v in object.items():
            if k == key:
                return path + [k]
            new_path = find_path(v, key, path + [k])
            if new_path:
                return new_path

    if isinstance(object, list):
        for idx, v in enumerate(object):
            new_path = find_path(v, key, path + [idx])
            if new_path:
                return new_path

    return None


def replace_using_path(object, path, new_value):
    current = object
    for idx, step in enumerate(path):
        if idx == len(path) - 1:
            current[step] = new_value
        elif isinstance(current, dict):
            current = current[step]
        elif isinstance(current, list):
            current = current[int(step)]
        else:
            raise ValueError("Invalid path or object structure")
    return object



class Aggregator:
    @classmethod
    def aggregate(cls, json_file: str, trace_dir: str):
        from pandas import read_json

        data_frame = read_json(json_file)
        cls.aggregate_data_frame(data_frame=data_frame, trace_dir=trace_dir)

    @classmethod
    def aggregate_data_frame(cls, data_frame, trace_dir):
        aggregated = data_frame.groupby("cohort_on").aggregate(list).transpose()
        with open(f"{trace_dir}/data.json", "w") as fp:
            fp.write(aggregated.to_json(default_handler=str))



from typing import Optional
from ..base.named_model import NamedModel
from abc import ABC, abstractmethod
from pydantic import Field, SecretStr


class DefaultTarget:
    pass


class Target(ABC, NamedModel):
    """
    Targets hold the connection information to your data sources.
    """

    host: Optional[str] = Field(None, description="The host url of the database.")
    port: Optional[int] = Field(None, description="The port of the database.")
    database: str = Field(
        ..., description="The database that the Visivo project will use in queries."
    )
    username: Optional[str] = Field(None, description="Username for the database.")
    password: Optional[SecretStr] = Field(
        None, description="Password corresponding to the username."
    )
    db_schema: Optional[str] = Field(
        None, description="The schema that the Visivo project will use in queries."
    )

    @abstractmethod
    def get_connection(self):
        raise NotImplementedError(f"No connection method implemented for {self.type}")

    @abstractmethod
    def read_sql(self, query: str):
        raise NotImplementedError(f"No read sql method implemented for {self.type}")

    def get_password(self):
        return self.password.get_secret_value() if self.password is not None else None

    def connect(self):
        return Connection(target=self)

    def url(self):
        from sqlalchemy.engine import URL

        url = URL.create(
            host=self.host,
            username=self.username,
            password=self.get_password(),
            port=self.port,
            drivername=self.get_dialect(),
            database=self.database,
            query=None,
        )
        return url


class Connection:
    def __init__(self, target: Target):
        self.target = target

    def __enter__(self):
        self.conn = self.target.get_connection()
        return self.conn

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.conn.close()
        self.conn = None
        self.conn_type = None



import time
import datetime
from dateutil import parser
import jinja2
import os
import json


def env_var(key):
    return os.getenv(key, "NOT-SET").replace("\\", "\\\\").replace('"', '\\"')


def now():
    return time.time()


def to_unix(date_str: str) -> float:
    """
    Accepts a string that represents a date or date & time and returns the UTC unix timestamp. Able to parse a wide range of
    different formats.

    The function attempts to be forgiving with regards to unlikely input formats,
    returning a datetime object even for dates which are ambiguous. If an element
    of a date/time stamp is omitted, the following rules are applied:

    - If AM or PM is left unspecified, a 24-hour clock is assumed, however, an hour
        on a 12-hour clock (``0 <= hour <= 12``) *must* be specified if AM or PM is
        specified.
    - If a time zone is omitted UTC is assumed.

    Args:
        date_str (str): The string representing the date or date & time.

    Returns:
        float: The UTC unix timestamp.
    """
    date_obj = parser.parse(date_str)

    if date_obj.tzinfo is None or date_obj.tzinfo.utcoffset(date_obj) is None:
        date_obj = date_obj.replace(tzinfo=datetime.timezone.utc)

    return date_obj.timestamp()


def to_iso(unix_timestamp: float):
    """
    Converts a UTC unix timestamp to an ISO 8601 formatted string.

    If the timestamp represents a date only (with no time component), the resulting
    string will be in the format 'YYYY-MM-DD'. Otherwise, the resulting string will
    be in the format 'YYYY-MM-DDTHH:MM:SSZ', where 'T' separates the date and time,
    and 'Z' indicates UTC time.

    Args:
        unix_timestamp (float): The UTC unix timestamp.

    Returns:
        str: The ISO 8601 formatted string.
    """
    date_obj = datetime.datetime.fromtimestamp(unix_timestamp, datetime.timezone.utc)
    if (
        date_obj.hour == 0
        and date_obj.minute == 0
        and date_obj.second == 0
        and date_obj.microsecond == 0
    ):
        return date_obj.date().isoformat()
    else:
        return date_obj.isoformat()


def to_str_format(unix_timestamp: float, str_format: str):
    """
    Converts a UTC unix timestamp to a string using the specified format.

    The format string should follow the directives of the Python `strftime` function.

    Args:
        unix_timestamp (float): The UTC unix timestamp.
        str_format (str): The format string.

    Returns:
        str: The formatted string.
    """
    date_obj = datetime.datetime.fromtimestamp(unix_timestamp, datetime.timezone.utc)
    return date_obj.strftime(str_format)


def read_json_file(filepath):
    """Read and parse a JSON file, returning a Python object.

    The filepath can be either relative to the script or module this function
    is defined in, or an absolute path.
    """
    # Check if the filepath is absolute. If not, treat it as relative to BASE_DIR.
    BASE_DIR = os.path.dirname(os.path.abspath(__file__))
    if not os.path.isabs(filepath):
        filepath = os.path.join(BASE_DIR, filepath)

    if not os.path.isfile(filepath):
        raise FileNotFoundError(f"File not found at: {filepath}")

    with open(filepath, "r") as file:
        data = json.load(file)
    return data


def timedelta(**kwargs):
    td = datetime.timedelta(**kwargs)
    return td.total_seconds()


FUNCTIONS = {
    "env_var": env_var,
    "now": now,
    "timedelta": timedelta,
    "to_unix": to_unix,
    "to_iso": to_iso,
    "to_str_format": to_str_format,
    "read_json_file": read_json_file,
}


def render_yaml(template_string: str):
    """
    Renders a YAML template string using Jinja2 and a set of predefined functions.

    Args:
        template_string (str): The YAML template string to render.

    Returns:
        str: The rendered YAML string.
    """
    template = jinja2.Template(template_string)
    return template.render(FUNCTIONS)



import os
import click
from pathlib import Path
from visivo.parsers.file_names import PROJECT_FILE_NAME
from visivo.commands.utils import get_profile_file
from visivo.utils import load_yaml_file


class Discover:
    def __init__(self, working_directory: str, home_directory=os.path.expanduser("~")):
        self.working_directory = working_directory
        self.home_directory = home_directory

    @property
    def project_file(self):
        return Path(self._get_any_project_file(self.working_directory))

    @property
    def files(self):
        if not os.path.exists(self.project_file):
            raise click.ClickException(f'Project file "{PROJECT_FILE_NAME}" not found')
        files = [self.project_file]

        self.__add_includes(files=files, file=self.project_file)

        profile_file = get_profile_file(home_directory=self.home_directory)
        if os.path.exists(profile_file):
            files.append(profile_file)

        return files

    def __add_includes(self, files, file):
        from visivo.models.include import Include

        data = load_yaml_file(file)
        base_path = os.path.dirname(file)

        if "includes" in data:
            for include_data in data["includes"]:
                include = Include(**include_data)
                include_path = f"{base_path}/{include.path}"
                if ".git" in include.path:
                    include_path = self.__get_project_file_from_git(
                        git_url=include.path
                    )

                if not os.path.exists(include_path):
                    raise click.ClickException(
                        f'Invalid "include" in project. "{include_path}" does not exist.'
                    )
                files.append(Path(include_path))
                self.__add_includes(files=files, file=include_path)

    def _get_any_project_file(self, dir):
        if os.path.exists(f"{dir}/visivo_project.yml"):
            return f"{dir}/visivo_project.yml"

        return f"{dir}/{PROJECT_FILE_NAME}"

    def __get_project_file_from_git(self, git_url):
        from git import Repo

        deps_folder = f"{self.working_directory}/.visivo_cache"
        if not os.path.exists(deps_folder):
            os.makedirs(deps_folder)
        if "@" not in git_url:
            raise click.ClickException(
                f'Invalid github dependency "{git_url}". A version specified with "@" is required'
            )

        repo_url = f"https://github.com/{git_url.split('@')[0]}"
        version = git_url.split("@")[1]
        if "--" in version:
            file = version.split("--")[1].strip()
            version = version.split("--")[0].strip()
        else:
            file = None

        local_folder = (
            f"{deps_folder}/{git_url.split('@')[0].replace('.git', '')}@{version}"
        )
        if os.path.exists(local_folder):
            repo = Repo(local_folder)
        else:
            repo = Repo.clone_from(repo_url, local_folder)

        try:
            repo.git.checkout(version)
        except Exception as e:
            repo.close()
            raise click.ClickException(f'Error cloning "{git_url}": {e}')

        if file:
            return f"{local_folder}/{file}"

        return self._get_any_project_file(local_folder)



from typing import List
from pydantic import Field
from visivo.models.base.base_model import generate_ref_field
from visivo.models.base.parent_model import ParentModel
from visivo.models.models.csv_script_model import CsvScriptModel
from visivo.models.models.model import Model
from visivo.models.targets.sqlite_target import Attachment, SqliteTarget
import os

from visivo.models.targets.target import Target


class LocalMergeModel(Model, ParentModel):
    """
    Local Merge Models are models that allow you to merge data from multiple other models locally.

    !!! note

        Any joining is done in a local SQLite database. It is not designed for large datasets.

    !!! example {% raw %}

        === "Internal join External"

            Here is an example of merging two models that are defined in your project. One that is external and one that is internal.
            ``` yaml
            models:
              - name: local_merge
                models:
                    - ref(first_domain_model)
                    - ref(external_data_model)
                sql: SELECT * FROM first_domain_model.model AS fdm JOIN external_data_model.model AS edm ON fdm.external_id = edm.id
            ```

    {% endraw %}
    """

    sql: str = Field(
        description="The sql used to generate your base data",
    )
    models: List[generate_ref_field(Model)] = Field(
        description="A model object defined inline or a ref() to a model."
    )

    def get_sqlite_target(self, output_dir, dag) -> SqliteTarget:
        attach = list(
            map(
                lambda model: Attachment(
                    schema_name=model.name,
                    target=self._get_sqlite_from_model(model, output_dir, dag),
                ),
                self._get_dereferenced_models(dag),
            )
        )

        return SqliteTarget(
            name=f"model_{self.name}_generated_target",
            database="",
            type="sqlite",
            attach=attach,
        )

    def insert_dependent_models_to_sqlite(self, output_dir, dag):
        for model in self._get_dereferenced_models(dag):
            if isinstance(model, CsvScriptModel):
                continue
            sqlite_target = self._get_sqlite_from_model(model, output_dir, dag)
            target = ParentModel.all_descendants_of_type(
                type=Target, dag=dag, from_node=model
            )[0]
            data_frame = target.read_sql(model.sql)
            engine = sqlite_target.get_engine()
            data_frame.to_sql("model", engine, if_exists="replace", index=False)

    def _get_sqlite_from_model(self, model, output_dir, dag) -> SqliteTarget:
        if isinstance(model, CsvScriptModel):
            return model.get_sqlite_target(output_dir=output_dir)
        elif isinstance(model, LocalMergeModel):
            return model.get_sqlite_target(output_dir=output_dir, dag=dag)
        else:
            return SqliteTarget(
                name=f"model_{model.name}_generated_target",
                database=f"{output_dir}/{model.name}.sqlite",
                type="sqlite",
            )

    def _get_dereferenced_models(self, dag):
        models = ParentModel.all_descendants_of_type(
            type=Model, dag=dag, from_node=self
        )
        return list(filter(lambda model: model is not self, models))

    def child_items(self):
        return self.models
